import os

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


# load calibration data
def load_calibration_data(calibration_data_filepath):
    """Load calibration data in a .npz file for the Reso spectrometer.

    Args:
        calibration_data_filepath (StrOrBytesPath): path of the file to load

    """
    try:
        calibration = np.load(calibration_data_filepath)
        wavelength_calibration = calibration["wavelength"]
        lamp_spec = calibration["lamp_ref"]
        lamp_measbyReso = calibration["lamp_measured"]
        calibration_smoothed = np.abs(calibration["corr_factor_smoothed"])
    except FileNotFoundError:
        print(
            "Error: calibration data for the UV spectrometer not found.\n"
            "Please copy the folder Attoworld/src/attoworld/spectrum/calibration_data into your current working directory\n"
            "or alternatively create a calibration file with relative path ./calibration_data/Reso_Spectrometer_CalibrationCorrection.npz\n"
        )
        raise FileNotFoundError("calibration data not found")
    return wavelength_calibration, lamp_spec, lamp_measbyReso, calibration_smoothed


def smooth(y, box_pts: int):
    """Basic box smoothing for the spectra.

    Arguments:
        y = spectrum to be smoothed [numpy array]
        box_pts = width of smoothing box [int, number of points]

    Returns:
        smoothed_y [numpy array]

    """
    box = np.ones(box_pts) / box_pts
    return np.convolve(y, box, mode="same")


def tukey_f(x: float, center: float, FWHM: float, w: float):
    """cosine-smoothed, flattop window function centered in 'center'.

    each edge is w-wide
    amplitude = 1

    Arguments:
        x: float = input value
        center = center of the t. function
        FWHM = full width at half maximum of the t. function
        w = width of the edges of the t. function (0 < w < FWHM)

    """
    if FWHM < w:
        print("tukey window can not have edges wider than FWHM")
    xmin = center - FWHM / 2 - w / 2
    xmax = center + FWHM / 2 + w / 2
    if x < xmin or x > xmax:
        y = 0
    elif xmin <= x and x < xmin + w:
        y = (1 - np.cos((x - xmin) * np.pi / w)) / 2
    elif xmin + w <= x and x <= xmax - w:
        y = 1
    elif xmax - w < x and x <= xmax:
        y = (1 - np.cos((x - xmax) * np.pi / w)) / 2
    else:
        print(x, xmax, xmin, w)
        raise ValueError(
            "in tukey_f, x could not be assigned correctly to the sub intervals, might it (or the other parameters) be NaN?"
        )
    return y


def tukey_window(x, center: float, FWHM: float, w: float):
    """Returns a cosine-smoothed, flattop window centered in 'center'; amplitude = 1.

    The difference with the function tukey_f is that this function can take as input a numpy array or a list, and returns a numpy array or a list.

    Args:
        x: input array (numpy array or list)
        center: center of the window
        FWHM: full width at half maximum
        w: width of the edges of the window (w > 0)

    Returns:
        window (numpy array or list)

    """
    if isinstance(x, np.ndarray):
        y = []
        for xi in x:
            y.append(tukey_f(xi, center, FWHM, w))
        y = np.array(y)
    elif isinstance(x, list):
        y = []
        for xi in x:
            y.append(tukey_f(xi, center, FWHM, w))
    elif isinstance(x, float):
        y = tukey_f(x, center, FWHM, w)
    else:
        raise TypeError(
            "in function tukey window x is neither np array nor list nor float"
        )
    return y


def read_csd_file(filename):
    """DEPRECATED read the file produced by the Maya spectrometer."""
    file = open(filename)
    data = []
    for row in file:
        # row = row.replace(',', '.')
        elemlist = str(row).split()
        row_float = []
        for elem in elemlist:
            row_float.append(float(elem))
        data.append(row_float)
    file.close()
    return np.array(data)


def plot_spectra(filenameList, pdfFilename, legendItemList=None):
    """DEPRECATED plot the spectra produced by the Maya spectrometer."""
    dataList = []
    for filename in filenameList:
        dataList.append(read_csd_file(filename))

    fig, ax = plt.subplots(2, 1, figsize=(6, 8))
    legendHandles = []
    for data in dataList:
        avgeSpectrum = []
        for row in data:
            avgeSpectrum.append(np.mean(np.array(row)))
        (handle,) = ax[0].plot(data[:, 0], avgeSpectrum)
        legendHandles.append(handle)
    # ax[0].set_ylim(3e-17,1e-9)
    # ax[0].set_xlim(freqLim[0], freqLim[1])
    ax[0].set_xlabel("Wavelength (nm)")
    ax[0].set_ylabel("Intensity")
    if legendItemList is not None:
        ax[0].legend(legendHandles, legendItemList)

    for data in dataList:
        avgeSpectrum = []
        for row in data:
            avgeSpectrum.append(np.mean(np.array(row)))
        (handle,) = ax[1].semilogy(data[:, 0], avgeSpectrum)
        legendHandles.append(handle)
    # ax[1].set_ylim(3e-17,1e-9)
    # ax[1].set_xlim(freqLim[0], freqLim[1])
    ax[1].set_xlabel("Wavelength (nm)")
    ax[1].set_ylabel("Intensity")
    if legendItemList is not None:
        ax[1].legend(legendHandles, legendItemList)

    fig.savefig(pdfFilename)


def read_spectrometer_excel(filename):
    """Reads xls file (passed as the string filename without extension!) produced by the UV-VIS spectrometer RESONANCE VS-7550-VUV.

    This function currently only works in Linux (Ubuntu), since it uses the OS system command to brutally copy the content of the xls file into a txt;
    such copying is necessary because the xls file is not readable by pandas (it's a non-standard xls, I couldn't find any other workaround).
    For windows there is a similar command, please replace the line
    os.system("cat " + filename + ".xls > " + filename + ".txt")

    44 (skipped) rows at the beginning of the file (all headers)

    Args:
        filename: name of the file (without extension)

    Returns:
        numpy array with the spectral data (see original excel for more details)

    """
    if "." in filename and "./" not in filename:
        raise ValueError(
            "in function read_spectrometer_excel, filename must be passed without extension (filename should not contain a dot)"
        )
    os.system("cat " + filename + ".xls > " + filename + ".txt")
    dataF = pd.read_table(
        filename + ".txt", sep="\t", keep_default_na=False, skiprows=44
    )  # keep_default_na=False,
    data = []
    for row in dataF.values:
        data.append([])
        for x in row:
            if x == "":
                data[-1].append(float("nan"))
            else:
                data[-1].append(float(x))
    return np.array(data)


def calibrate(
    data,
    column: int,
    calibration_file_path="./calibration_data/Reso_Spectrometer_CalibrationCorrection.npz",
    dark=None,
    dark_c=None,
    stitch: bool = False,
    smooth_points: int = 10,
    null_calibration: bool = False,
    wavelength_calibration_intercept: float = 3.538,
    wavelength_calibration_slope: float = 1.003427,
):
    """For the UV spectrometer. Calibrate the spectrum number 'column' in the the 'data' array. Intensity calibration factor is loaded in the first lines of UVSpectrumAnalysis.py.
        Notice: each saved spectrum has 7 columns in the excel file; however, the argument column refers to the index of the spectrum, not the actual column in the file.

    Args:
        data: data [numpy] array (see read_spectrometer_excel)
        column: index of the spectrum to be calibrated
        calibration_file_path: path of the calibration file
        dark: dark spectrum (optional)
        dark_c: column of the dark spectrum (optional)
        stitch: Boolean, if True, the spectra are stitched together (optional, default False)
                if True, the spectrum will be calculated by stitching the 5 partial spectra in the previous columns.
                the stitching only works if the spectrum number 'column' is a FULL-SPECTRUM in the xls file.
                Stitching was implemented because the internal software of the spectrometer is not good enough at it (the saved full spectrum typically has discontinuities).
        smooth_points: number of points for smoothing (optional, default 10). if 0, no smoothing is applied
        null_calibration: Boolean, if True, the intensity calibration is not applied (optional, default False)
        wavelength_calibration_intercept: intercept of the wavelength calibration (optional, default 3.538). If None no wavelength calibration is applied
        wavelength_calibration_slope: slope of the wavelength calibration (optional, default 1.003427). If None no wavelength calibration is applied
                Wavelength calibration is in the form λ_true = λ_measured * slope + intercept
                Notice: wavelength calibration has an error of +/- 1  to 2 nm due to the stitching process.
                    for a better calibration compute and apply the correction to each of the 5 partial spectra

    Returns:
        wavelength: wavelength array
        spectrum: calibrated spectrum

    """
    wavelength_calibration, lamp_spec, lamp_measbyReso, calibration_smoothed = (
        load_calibration_data(calibration_file_path)
    )

    if wavelength_calibration_slope is None or wavelength_calibration_intercept is None:
        cal_slope = 1.0
        cal_intercept = 0.0
    else:
        cal_slope = wavelength_calibration_slope
        cal_intercept = wavelength_calibration_intercept

    if stitch:
        # check that the chosen spectrum (in column) is actually a full one
        theoretically_complete_wvl = (
            data[:, 7 * (column) + 1] * cal_slope + cal_intercept
        )
        theoretically_complete_wvl = theoretically_complete_wvl[
            ~np.isnan(theoretically_complete_wvl)
        ]
        if theoretically_complete_wvl[-1] - theoretically_complete_wvl[0] < 900:
            raise ValueError(
                "in function calibrate(), the spectrum to be stitched is not a full spectrum, please check the column index"
            )

        extrema = []
        spectra = []
        for i in range(5):
            wvl_single_sp = (
                data[:, 7 * (column - 5 + i) + 1] * cal_slope + cal_intercept
            )
            single_sp = data[:, 7 * (column - 5 + i) + 2]
            if dark is not None and dark_c is not None:
                single_sp -= dark[:, 7 * (dark_c - 5 + i) + 2]
            wvl_single_sp = wvl_single_sp[~np.isnan(wvl_single_sp)]
            single_sp = single_sp[~np.isnan(single_sp)]
            extrema.append([wvl_single_sp[0], wvl_single_sp[-1]])
            spectra.append([wvl_single_sp, single_sp])

        overlaps = []
        centers_of_overlap = [extrema[0][0]]
        for i in range(4):
            overlaps.append(extrema[i][1] - extrema[i + 1][0])
            centers_of_overlap.append(extrema[i][1] / 2 + extrema[i + 1][0] / 2)

        w = np.min(overlaps)
        centers_of_overlap.append(extrema[4][1] + w / 2)
        centers_of_overlap[0] = centers_of_overlap[0] - w / 2

        wavelength = data[:, 7 * column + 1] * cal_slope + cal_intercept
        spectrum = wavelength * 0.0

        for i in range(5):
            center = centers_of_overlap[i] / 2 + centers_of_overlap[i + 1] / 2
            FWHM = centers_of_overlap[i + 1] - centers_of_overlap[i]
            spectrum += np.interp(
                wavelength,
                spectra[i][0],
                spectra[i][1] * tukey_window(spectra[i][0], center, FWHM, w),
            )
            plt.plot(
                spectra[i][0],
                spectra[i][1] * tukey_window(spectra[i][0], center, FWHM, w),
            )
        calibr_interpolated = np.interp(
            wavelength, wavelength_calibration, calibration_smoothed
        )
        if not null_calibration:
            spectrum *= calibr_interpolated

        plt.plot(
            data[:, 7 * column + 1] * cal_slope + cal_intercept, data[:, 7 * column + 2]
        )
        plt.plot(wavelength, spectrum)

        if w < 0:
            raise ValueError(
                "in function calibrate(), the spectra to be stitched do not overlap"
            )

    else:
        wavelength = data[:, 7 * column + 1] * cal_slope + cal_intercept
        if dark is None or dark_c is None:
            calibr_interpolated = np.interp(
                wavelength, wavelength_calibration, calibration_smoothed
            )
            if null_calibration:
                calibr_interpolated = calibr_interpolated * 0.0 + 1
            spectrum = (
                data[:, 7 * column + 2] - np.min(data[:, 7 * column + 2])
            ) * calibr_interpolated
        else:
            calibr_interpolated = np.interp(
                wavelength, wavelength_calibration, calibration_smoothed
            )
            if null_calibration:
                calibr_interpolated = calibr_interpolated * 0.0 + 1
            spectrum = (
                data[:, 7 * column + 2] - dark[:, 7 * dark_c + 2]
            ) * calibr_interpolated

            if len(data[:, 7 * column + 2][~np.isnan(data[:, 7 * column + 2])]) != len(
                dark[:, 7 * dark_c + 2][~np.isnan(dark[:, 7 * dark_c + 2])]
            ):
                raise ValueError(
                    "In function calibrate(), the dark spectrum is not the same length as the spectrum to be calibrated.\n"
                    "Please check the column indices of the dark spectrum and the main spectrum"
                )
            if (
                data[:, 7 * column + 1][~np.isnan(data[:, 7 * column + 1])][0]
                != dark[:, 7 * dark_c + 2][~np.isnan(dark[:, 7 * dark_c + 1])][0]
            ):
                print(
                    "WARNING: In function calibrate(), the dark spectrum does not have the same wavelength as the spectrum to be calibrated.\n"
                    "Please check the column indices of the dark spectrum and the main spectrum"
                )

    if smooth_points is not None and smooth_points > 0:
        spectrum = smooth(spectrum, smooth_points)
    return wavelength, spectrum


def list_error():
    raise Exception("list size in function plot_spectra_UVsp is not coherent")


def plot_spectra_UVsp(
    filenameList,
    columnList,
    pdfFilename,
    legendItemList=None,
    darkTupleList=None,
    normalizationList=None,
    color_gradient: bool = False,
    additive_constant_log_intensity=20,
    wavelength_range=None,
    title=None,
    invert_order: bool = False,
    plotList=None,
    do_calibrate: bool = True,
    stitch: bool = False,
    smooth_points: int = 10,
    calibration_file_path="./calibration_data/Reso_Spectrometer_CalibrationCorrection.npz",
):
    """Plotting function for the UV spectrometer.

    Notice: The excel file usually contains several spectra.
    Notice: each saved spectrum has 7 columns in the excel file; please consider that the argument 'column' refers to the index of the spectrum, not the actual column in the file.

    Args:
        filenameList: input file list (without extension)
        columnList: list of the positions in the file of the spectra to be plotted (e.g. [1, 0, 5] -> plot 1st sp. for 1st file, 5th spectrum for 3rd file, ...)
        pdfFilename: output filename
        darkTupleList: [("darkFilename1stspectrum", spectrum_number_in_1stDark_filename), ( ..,.. ), ...]
        legendItem List: list of entries for the legend (optional, keyword)
        normalizationList: list of (multiplicative) normalization factors for the spectra (optional, keyword)
        color_gradient: Boolean, whether plot should be displayed with color gradient instead of in matplotlib default color cycle
        additive_constant_log_intensity: additive constant to the spectral intensity, so that the logarithmic plot does not have values too close to 0
        wavelength_range: [min lambda, max lambda] for plotting
        title: title of the figure
        plotList: list of booleans: this function will plot the entry filenameList[i] if plotList[i] = True, otherwise the i-th data will not be shown in the plot
        invert_order: Boolean, if True, the order of the spectra will be inverted (i.e. the last spectrum in the list will be plotted first)
        do_calibrate: Boolean, if True, the spectra will be calibrated (default True, see calibrate() function)
        calibration_file_path: str, path to the calibration file (default "./calibration_data/Reso_Spectrometer_CalibrationCorrection.npz"). Ignored if do_calibrate = False
        stitch: Boolean, if True, the spectra will be stitched together (optional, default False)
            if True, the spectrum will be calculated by stitching the 5 partial spectra in the previous columns.
            the stitching only works if the spectrum number 'column' is a FULL-SPECTRUM in the xls file.
            Stitching was implemented because the internal software of the spectrometer does not do it properly (the saved full spectrum tipically has discontinuities).
            Stitching here works only if do_calibrate = True
        smooth_points: number of points for smoothing (optional, default 10)

    DEPENDS ON read_spectrometer_excel()
    DEPENDS ON calibrate() [optional]

    """
    if len(columnList) != len(filenameList):
        print(
            "Error: in function plot_spectra_UVsp\ncolumnList has different size than filenameList"
        )
        list_error()
    if legendItemList is not None:
        if len(legendItemList) != len(filenameList):
            print(
                "Error: in function plot_spectra_UVsp\nlegendItemList has different size than filenameList"
            )
            list_error()
    if darkTupleList is not None:
        if len(darkTupleList) != len(filenameList):
            print(
                "Error: in function plot_spectra_UVsp\ndarkTupleList has different size than filenameList"
            )
            list_error()
    if normalizationList is not None:
        if len(normalizationList) != len(filenameList):
            print(
                "Error: in function plot_spectra_UVsp\nnormalizationList has different size than filenameList"
            )
            list_error()
    if plotList is not None:
        if len(plotList) != len(filenameList):
            print(
                "Error: in function plot_spectra_UVsp\nplotList has different size than filenameList"
            )
            list_error()

    if invert_order:
        filenameList.reverse()
        columnList.reverse()
        if legendItemList is not None:
            legendItemList.reverse()
        if darkTupleList is not None:
            darkTupleList.reverse()
        if normalizationList is not None:
            normalizationList.reverse()
        if plotList is not None:
            plotList.reverse()

    if plotList is not None:
        for i in reversed(range(len(plotList))):
            if not plotList[i]:
                if filenameList is not None:
                    del filenameList[i]
                if columnList is not None:
                    del columnList[i]
                if legendItemList is not None:
                    del legendItemList[i]
                if darkTupleList is not None:
                    del darkTupleList[i]
                if normalizationList is not None:
                    del normalizationList[i]

    dataList = []
    for filename in filenameList:
        data = read_spectrometer_excel(filename)
        dataList.append(data)

    # check if the dark spectrum is shorter or longer than the spectrum to be calibrated (if it has nan values)
    # [when a = nan the statement (a == a) returns False]
    if darkTupleList is not None:
        for tp, data, column in zip(darkTupleList, dataList, columnList):
            dark = read_spectrometer_excel(tp[0])
            for i in range(len(data[:, 7 * column + 2])):
                if (
                    dark[i, 7 * tp[1] + 2] != dark[i, 7 * tp[1] + 2]
                    and data[i, 7 * column + 2] == data[i, 7 * column + 2]
                ):
                    print(
                        "Warning: in function plot_spectra_UVsp\ndark spectrum is shorter than the spectrum to be calibrated"
                        "\nplease check the column index of the dark spectrum"
                    )
                    break
                if (
                    data[i, 7 * column + 2] != data[i, 7 * column + 2]
                    and dark[i, 7 * tp[1] + 2] == dark[i, 7 * tp[1] + 2]
                ):
                    print(
                        "Warning: in function plot_spectra_UVsp\ndark spectrum is longer than the spectrum to be calibrated"
                        "\nplease check the column indices of the dark spectrum and the main spectrum"
                    )
                    break

    # subtract dark (if any, and if it will not be subtracted during calibration)
    if darkTupleList is not None and not do_calibrate:
        for tp, data, column in zip(darkTupleList, dataList, columnList):
            dark = read_spectrometer_excel(tp[0])
            for i in range(len(data[:, 7 * column + 2])):
                data[i, 7 * column + 2] = (
                    data[i, 7 * column + 2] - dark[i, 7 * tp[1] + 2]
                )

    if do_calibrate:
        for i in range(len(dataList)):
            if darkTupleList is not None:
                dark = read_spectrometer_excel(darkTupleList[i][0])
                (
                    dataList[i][:, 7 * columnList[i] + 1],
                    dataList[i][:, 7 * columnList[i] + 2],
                ) = calibrate(
                    dataList[i],
                    columnList[i],
                    calibration_file_path=calibration_file_path,
                    dark=dark,
                    dark_c=darkTupleList[i][1],
                    stitch=stitch,
                    smooth_points=smooth_points,
                )
            else:
                (
                    dataList[i][:, 7 * columnList[i] + 1],
                    dataList[i][:, 7 * columnList[i] + 2],
                ) = calibrate(
                    dataList[i],
                    columnList[i],
                    calibration_file_path=calibration_file_path,
                    calibration_dark=None,
                    dark_c=None,
                    stitch=stitch,
                    smooth_points=smooth_points,
                )

    n_lines = len(dataList)
    if color_gradient:
        # Take colors at regular intervals spanning the colormap.
        cmap = mpl.colormaps[
            "magma"
        ]  #'viridis', 'plasma', 'inferno', 'magma', 'cividis'
        colors = cmap(np.linspace(0, 0.92, n_lines))
        # fig, ax = plt.subplots(layout='constrained')
    else:
        prop_cycle = plt.rcParams["axes.prop_cycle"]
        colors = prop_cycle.by_key()["color"]
        # colors = colors[:n_lines]

    fig, ax = plt.subplots(2, 1, figsize=(6, 8))
    if title is not None:
        fig.suptitle(title)
    ax[0].xaxis.set_minor_locator(mpl.ticker.AutoMinorLocator())
    ax[1].xaxis.set_minor_locator(mpl.ticker.AutoMinorLocator())
    legendHandles = []
    if normalizationList is None:
        normalizationList = [1.0 for i in dataList]
    for data, column, normalization, color, i in zip(
        dataList, columnList, normalizationList, colors, range(len(dataList))
    ):
        (handle,) = ax[0].plot(
            data[:, 7 * column + 1],
            data[:, 7 * column + 2] * normalization,
            color=color,
        )
        legendHandles.append(handle)
    if wavelength_range is not None:
        ax[0].set_xlim(wavelength_range[0], wavelength_range[1])
    ax[0].set_xlabel("Wavelength (nm)")
    ax[0].set_ylabel("Intensity")
    ax[0].grid(linestyle=(0, (5, 10)))  # offset, dash length, space length
    if legendItemList is not None:
        ax[0].legend(legendHandles, legendItemList)

    for data, column, normalization, color, i in zip(
        dataList, columnList, normalizationList, colors, range(len(dataList))
    ):
        (handle,) = ax[1].semilogy(
            data[:, 7 * column + 1],
            data[:, 7 * column + 2] * normalization + additive_constant_log_intensity,
            color=color,
        )
        legendHandles.append(handle)
    if wavelength_range is not None:
        ax[1].set_xlim(wavelength_range[0], wavelength_range[1])
    ax[1].set_xlabel("Wavelength (nm)")
    ax[1].set_ylabel("Intensity")
    ax[1].grid(linestyle=(0, (5, 10)))
    if legendItemList is not None:
        ax[0].legend(legendHandles, legendItemList)

    fig.savefig(pdfFilename)
    return fig



"""Tools for handling the various data types used across the labs."""

from .decorators import add_method, yaml_io
from .file_io import (
    load_mean_spectrum_from_scarab,
    load_spectrum_from_text,
    load_waveform_from_text,
    load_waves_from_matfile,
    read_dwc,
    read_Trebino_FROG_data,
    read_Trebino_FROG_matrix,
    read_Trebino_FROG_speck,
)
from .frog_data import FrogData, Spectrogram
from .interop import (
    ComplexEnvelope,
    ComplexSpectrum,
    IntensitySpectrum,
    Waveform,
)
from .luna_result import LunaResult
from .spectrogram import FrogBinSettings
from .spectrometer_calibration_dataset import (
    CalibrationDataset,
    CalibrationInput,
    SpectrometerCalibration,
)

__all__ = [
    "read_dwc",
    "load_mean_spectrum_from_scarab",
    "load_waves_from_matfile",
    "load_waveform_from_text",
    "load_spectrum_from_text",
    "FrogData",
    "read_Trebino_FROG_matrix",
    "read_Trebino_FROG_speck",
    "read_Trebino_FROG_data",
    "LunaResult",
    "yaml_io",
    "add_method",
    "SpectrometerCalibration",
    "Waveform",
    "ComplexSpectrum",
    "IntensitySpectrum",
    "ComplexEnvelope",
    "Spectrogram",
    "FrogBinSettings",
    "CalibrationDataset",
    "CalibrationInput",
]



import marimo

__generated_with = "0.14.10"
app = marimo.App(width="medium")


@app.cell
def _():
    import timeit

    import attoworld as aw
    import marimo as mo
    import matplotlib.pyplot as plt
    import numpy as np
    import scipy

    aw.plot.set_style("nick_dark", font_size=14)
    return aw, mo, np, plt, scipy, timeit


@app.cell
def _(mo):
    mo.md(r"""## Check convergence of numerical derivatives""")
    return


@app.cell
def _(aw, np, plt):
    def convergence_check(order, N_pts):
        x = np.linspace(0.0, 2 * np.pi, N_pts + 1)[0:-1]
        dx = x[1] - x[0]
        y = np.cos(x)
        y_derivative_analytic = -np.sin(x)
        y_derivative = aw.numeric.derivative(y, 1, order) / dx
        # y_derivative = aw.numeric.derivative_periodic(y,1,order)/dx
        return np.max(np.abs(y_derivative - y_derivative_analytic))

    N_pts_range = [16, 24, 32, 48, 64, 96, 128, 192, 256, 384, 512]
    order_range = range(1, 6)

    convergence_data = np.array(
        [
            [convergence_check(_order, _n) for _n in N_pts_range]
            for _order in order_range
        ]
    )

    lines = plt.loglog(N_pts_range, convergence_data.T)
    for order, line in zip(order_range, lines):
        line.set_label(f"{order}")
    plt.xlabel("Number of grid points")
    plt.ylabel("Mean error")
    plt.ylim(1e-15, 0.1)
    plt.legend()
    aw.plot.showmo()
    return (N_pts_range,)


@app.cell
def _(mo):
    mo.md(
        r"""
    ## Check interpolate()
    The interpolate function uses Fornberg's algorithm to generate interpolation stencils. It should have performance and accuracy similar to scipy's CubicSpline when neighbors=3. It will be comparatively slower if you would have reused the same CubicSpline object for multiple interpolations, however.
    """
    )
    return


@app.cell
def _(aw, np, plt, scipy):
    def plot_interpolate_test():
        def test_function(_x):
            beta = 11.0
            return np.sin(_x**2 / beta)

        x = np.real(np.linspace(0.0, 16.0, 64))
        x2 = np.linspace(0.01, 15.99, 333)
        y = test_function(x)
        y2 = test_function(x2)

        interpolated_rs = aw.numeric.interpolate(x2, x, y, neighbors=3)
        interpolated_scipy = scipy.interpolate.CubicSpline(x, y)(x2)

        fig, ax = plt.subplots(2, 1, figsize=(6, 6))
        ax[0].plot(x2, interpolated_rs, label="aw.numeric.interpolate")
        ax[0].plot(x2, interpolated_scipy, "--", label="scipy.interpolate.CubicSpline")
        ax[0].plot(x, y, ".", label="Input data")
        ax[0].set_xlabel("x")
        ax[0].set_ylabel("y")
        ax[0].legend(loc="lower left")
        aw.plot.label_letter("a", axis=ax[0])
        ax[1].semilogy(x2, np.abs(y2 - interpolated_rs), label="aw.numeric.interpolate")
        ax[1].plot(
            x2, np.abs(y2 - interpolated_scipy), label="scipy.interpolate.CubicSpline"
        )
        ax[1].set_xlabel("x")
        ax[1].set_ylabel("Error")
        ax[1].legend()
        aw.plot.label_letter("b", axis=ax[1])

    plot_interpolate_test()
    aw.plot.showmo()
    return


@app.cell
def _(aw, np, scipy, timeit):
    def benchmark_interpolations():
        beta = 9.0
        x = np.real(np.linspace(0.0, 16.0, 333))
        x2 = np.linspace(0.1, 15.99, 2222)
        y = np.sin(x**2 / beta)
        print(
            f"aw.numeric.interpolate: {timeit.timeit(lambda: aw.numeric.interpolate(x2, x, y, neighbors=3), number=1000)} seconds"
        )
        print(
            f"scipy.interpolate.CubicSpline: {timeit.timeit(lambda: scipy.interpolate.CubicSpline(x, y, extrapolate=False)(x2), number=1000)} seconds"
        )

    benchmark_interpolations()
    return


@app.cell
def _(mo):
    mo.md(
        r"""
    ## Check FWHM error
    Note that the aw.numeric.fwhm() function has a much smaller error than the implementation I've seen elsewhere (which only reduces in error linearly with step size).
    """
    )
    return


@app.cell
def _(N_pts_range, aw, np, plt):
    def plot_fwhm_convergence():
        true_fwhm = 2 * np.sqrt(2.0 * np.log(2.0))
        fwhm = []
        fwhm_th = []
        dx_array = []

        def old_fwhm(y, x):
            i_max = np.argmax(y)
            hm = y[i_max] / 2
            return np.max(x[np.argwhere(y >= hm)]) - np.min(x[np.argwhere(y >= hm)])

        for i in range(len(N_pts_range)):
            x = np.linspace(-5, 5, N_pts_range[i])
            dx = x[1] - x[0]
            y = np.exp(-(x**2) / 2)
            fwhm.append(aw.attoworld_rs.fwhm(y, dx, neighbors=2))
            fwhm_th.append(old_fwhm(y, x))
            dx_array.append(dx)
        plt.loglog(N_pts_range, dx_array, label="time step")
        plt.loglog(
            N_pts_range, np.abs(fwhm - true_fwhm), label="FWHM in attoworld.numeric"
        )
        plt.loglog(N_pts_range, np.abs(fwhm_th - true_fwhm), label="index-based FWHM")
        plt.xlabel("# of points")
        plt.ylabel("Error")
        plt.legend()

    plot_fwhm_convergence()
    aw.plot.showmo()
    return


@app.cell
def _(mo):
    mo.md(
        r"""
    ## Labels
    The attoworld.plot.Char class contains useful symbols for putting correct labels on things.
    """
    )
    return


@app.cell
def _(aw, np, plt):
    lam = np.linspace(100.0, 1000.0, 1024)
    signal = np.random.rand(lam.shape[0])
    plt.plot(lam, signal)
    plt.xlabel(aw.plot.Char.wavelength_micron)
    plt.ylabel("Intensity (Arb. unit)")
    aw.plot.showmo()
    return


@app.cell
def _(mo):
    mo.md(r"""## Data containers""")
    return


@app.cell
def _(aw, np, plt):
    _t = 1e-15 * np.linspace(-50.0, 50.0, 2048)
    _E = np.exp(-(_t**2) / (2 * 50e-15**2)) * np.cos(2 * np.pi * 400e12 * _t)
    _w = aw.data.Waveform(
        wave=_E, time=_t, dt=(_t[1] - _t[0]), is_uniformly_spaced=True
    )
    _s = _w.to_intensity_spectrum()
    _fig, _ax = plt.subplots(2, 1)
    _ax[0].plot(1e15 * _w.time, _w.wave)
    _ax[0].plot(1e15 * _w.time, _w.to_windowed("tukey").wave)
    _ax[0].plot(
        1e15 * _w.time, _w.to_windowed("tukey").to_bandpassed(400e12, 20e12, 4).wave
    )
    _ax[0].set_xlabel("Time (fs)")
    _ax[1].loglog(1e6 * _s.wavelength, _s.spectrum)
    _ax[1].loglog(
        1e6 * _s.wavelength,
        1e9 * _w.to_intensity_spectrum(wavelength_scaled=False).spectrum,
    )
    _ax[1].loglog(
        1e6 * _s.wavelength, _w.to_windowed("tukey").to_intensity_spectrum().spectrum
    )
    _ax[1].loglog(
        1e6 * _s.wavelength, _w.to_windowed("blackman").to_intensity_spectrum().spectrum
    )
    _ax[1].set_ylim(1, 1e18)
    _ax[1].set_xlabel(aw.plot.Char.wavelength_micron)
    aw.plot.showmo()
    return


if __name__ == "__main__":
    app.run()



"""Class for handling a complex pulse envelope."""

import copy
from dataclasses import dataclass
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes

from ..numeric import derivative, fwhm
from .decorators import yaml_io


@yaml_io
@dataclass(slots=True)
class ComplexEnvelope:
    """Data corresponding to a complex envelope of a pulse, e.g. from a FROG measurement.

    Attributes:
        envelope (np.ndarray): the complex envelope
        time: (np.ndarray): the time array
        dt (float): the time step
        carrier_frequency (float): the carrier frequency of the envelope

    """

    envelope: np.ndarray
    time: np.ndarray
    dt: float
    carrier_frequency: float = 0.0

    def lock(self):
        """Make the data immutable."""
        self.envelope.setflags(write=False)
        self.time.setflags(write=False)

    def time_fs(self):
        """Time axis in femtoseconds."""
        return 1e15 * self.time

    def copy(self):
        """Return a copy."""
        return copy.deepcopy(self)

    def get_fwhm(self) -> float:
        """Full-width-at-half-maximum value of the envelope
        Returns:
            float: the fwhm.
        """
        return fwhm(np.abs(self.envelope) ** 2, self.dt)

    def plot(self, ax: Optional[Axes] = None, phase_blanking: float = 0.05, xlim=None):
        """Plot the pulse.

        Args:
            ax: optionally plot onto a pre-existing matplotlib Axes
            phase_blanking: only show phase information (instantaneous frequency) above this level relative to max intensity
            xlim: pass arguments to set_xlim() to constrain the x-axis

        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()
        time_ax = self.time_fs() - np.mean(self.time_fs())
        intensity = np.abs(self.envelope) ** 2
        intensity /= np.max(intensity)
        intensity_line = ax.plot(
            time_ax,
            intensity,
            label=f"Intensity, fwhm {1e15 * self.get_fwhm():0.1f} fs",
        )
        ax.set_xlabel("Time (fs)")
        ax.set_ylabel("Intensity (Arb. unit)")
        inst_freq = (
            (1e-12 / (2 * np.pi))
            * derivative(np.unwrap(np.angle(self.envelope)), 1)
            / self.dt
        )
        ax_phase = plt.twinx(ax)
        assert isinstance(ax_phase, Axes)
        ax_phase.plot([], [])
        phase_line = ax_phase.plot(
            time_ax[intensity > phase_blanking],
            inst_freq[intensity > phase_blanking],
            "--",
            label="Inst. frequency",
        )
        ax_phase.set_ylabel("Inst. frequency (THz)")
        if xlim is not None:
            ax.set_xlim(xlim)
            ax_phase.set_xlim(xlim)
        lines = lines = intensity_line + phase_line
        ax.legend(lines, [str(line.get_label()) for line in lines])
        return fig



"""Define the base class for handling spectrometer calibrations."""

from dataclasses import dataclass

import numpy as np

from .. import spectrum
from .decorators import yaml_io


@yaml_io
@dataclass
class SpectrometerCalibration:
    """Set of data describing a spectrometer calibration.

    Attributes:
        intensity_factors (np.ndarray): the intensity correction factors (weights)
        corrected_wavelengths (np.ndarray): the corrected wavelength array of the spectrometer
        corrected_frequencies (np.ndarray): the corrected frequency array of the spectrometer (c/wavelengths)

    """

    intensity_factors: np.ndarray
    original_wavelengths: np.ndarray
    corrected_wavelengths: np.ndarray
    corrected_frequencies: np.ndarray

    def save_npz(self, filepath):
        """Save to an npz file.

        Args:
            filepath: path to save to

        """
        np.savez(
            filepath,
            intensity_factors=self.intensity_factors,
            original_wavelengths=self.original_wavelengths,
            corrected_wavelengths=self.corrected_wavelengths,
            corrected_frequencies=self.corrected_frequencies,
        )

    @staticmethod
    def from_npz(filepath):
        """Make an instance of the class from an npz file.

        Args:
            filepath: path to the file

        Returns:
            SpectrometerCalibration: the calibration in the file

        """
        npzfile = np.load(filepath)
        return SpectrometerCalibration(
            intensity_factors=npzfile["intensity_factors"],
            original_wavelengths=npzfile["original_wavelengths"],
            corrected_wavelengths=npzfile["corrected_wavelengths"],
            corrected_frequencies=npzfile["corrected_frequencies"],
        )

    @staticmethod
    def from_named(spectrometer: spectrum.CalibrationData):
        """Loads a calibration saved in the database.

        Args:
            spectrometer (spectrum.CalibrationData): Value from the CalibrationData enum attoworld.spectrum.CalibrationData

        Returns:
            SpectrometerCalibration: the calibration associated with the enum value

        """
        return SpectrometerCalibration.from_npz(
            spectrum.get_calibration_path() / spectrometer.value
        )



"""Data processing routines that operate on measured or simulated waveforms."""

from .frog import (
    bundle_frog_reconstruction,
    generate_spectrogram,
    reconstruct_blindfrog,
    reconstruct_frog,
    reconstruct_xfrog,
)
from .wave import align_waves

__all__ = [
    "align_waves",
    "reconstruct_frog",
    "reconstruct_xfrog",
    "reconstruct_blindfrog",
    "generate_spectrogram",
    "bundle_frog_reconstruction",
]



"""Tools from the Attosecond science group at the Max Planck Institute of Quantum Optics, a.k.a. [Attoworld](https://www.attoworld.de).

The module has several submodules to keep it organized:
- **[data](attoworld/data.html)**: classes and for handling the various data formats used across the labs
- **[numeric](attoworld/numeric.html)**: numerical tools
- **[personal](attoworld/personal.html)**: a module where we can add our own functions that might not be of general interest, but are still good to have available to we can easily share work
- **[plot](attoworld/plot.html)**: functions for plotting with a consistent style
- **[wave](attoworld/wave.html)**: functions for processing waveforms and pulses
- **[spectrum](attoworld/spectrum.html)**: functions for processing spectra
- **[attoworld_rs](attoworld/attoworld_rs.html)**: A place to put Rust code with a Python interface for cases where it's particularly important that the program be fast and correct.
"""

from . import attoworld_rs, data, numeric, personal, plot, spectrum, wave

__all__ = ["data", "numeric", "personal", "plot", "spectrum", "wave", "attoworld_rs"]



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.ndimage import gaussian_filter1d


def box_smooth(y, box_pts):
    box = np.ones(box_pts) / box_pts
    return np.convolve(y, box, mode="same")


def eliminate_outliers(y, threshold: float = 3, window_points: int = 20):
    """Eliminates outliers in the data by replacing them with the mean of the surrounding values.

    Args:
        y (np.ndarray): Input data array.
        threshold (float): Threshold in units of sigma for outlier detection. Default is 3.
        window_points (int): Number of points to consider for the mean and sigma calculation. Default is 20.

    """
    if not isinstance(y, np.ndarray):
        raise TypeError("Input of eliminate_outliers must be a numpy array.")
    n_outliers = 0
    for i in range(len(y)):
        window_lower = int(max(0, i - window_points / 2))
        window_upper = int(min(len(y), i + window_points / 2))
        mean = np.mean(y[window_lower:window_upper])
        std = np.std(y[window_lower:window_upper])
        if abs(y[i] - mean) > threshold * std:
            y[i] = np.mean(np.concatenate((y[window_lower:i], y[i + 1 : window_upper])))
            n_outliers += 1

    print(
        n_outliers, " outliers detected and replaced with mean of surrounding values."
    )


def read_spectrum_maya(
    filename,
    remove_offsets_individually=False,
    nm_smearing=1.0,
    eliminate_outliers_spectrum=False,
):
    """Reads a spectrum file from the Maya spectrometer and returns the wavelength and spectrum arrays.

    The file is expected to have one column of wavelength data and arbitrarily many columns of equivalent spectra [to be averaged].
    The function also applies a Gaussian filter to the spectrum data to smooth it.
    The function can optionally eliminate outliers in the spectrum data and remove offsets from each column individually.

    Args:
        filename
        remove_offsets_individually (bool): If True, compute and remove offsets from each column individually
        nm_smearing (float): sigma in nm of the gaussian filter applied to smooth the spectrum.
        eliminate_outliers_spectrum (bool): = If True, eliminate single pixel outliers in the spectral data and replace them with average of surrounding values.
            default is False.

    """
    data = pd.read_table(filename, sep=" ", keep_default_na=True, skiprows=0)
    spectrum = []
    offsets = [0.0] * len(data.columns)
    fig, ax = plt.subplots()
    wvl = np.array(data.iloc[:, 0])
    final_nm_smearing = nm_smearing
    if eliminate_outliers_spectrum:
        nm_smearing = 0
    for i in range(len(data.columns) - 1):
        if nm_smearing > 0:
            spectrum_i = gaussian_filter1d(
                np.array(data.iloc[:, i + 1]), sigma=nm_smearing / (wvl[1] - wvl[0])
            )
        else:
            spectrum_i = np.array(data.iloc[:, i + 1])
        if eliminate_outliers_spectrum:
            eliminate_outliers(spectrum_i, threshold=3, window_points=20)
            spectrum_i = gaussian_filter1d(
                spectrum_i, sigma=final_nm_smearing / (wvl[1] - wvl[0])
            )
        offsets[i] = spectrum_i.min()
        if remove_offsets_individually:
            spectrum.append(spectrum_i - offsets[i])
            ax.plot(wvl, spectrum_i - offsets[i], label=f"Channel {i + 1}")
        else:
            spectrum.append(spectrum_i)
            ax.plot(wvl, spectrum_i, label=f"Channel {i + 1}")

    spectrum = np.mean(np.array(spectrum), axis=0)
    ax.plot(wvl, spectrum, label="Mean spectrum")

    return wvl, np.maximum(np.zeros(len(spectrum)), spectrum)


def read_spectrum_ocean_optics(filename):
    """Reads a spectrum file from the Ocean Optics spectrometer and returns the wavelength and spectrum arrays.
    The file is expected to have one column of wavelength data and one column of spectrum data.
    wavelengths are in nm.
    14 lines of header are skipped.
    """
    data = pd.read_table(filename, sep="\t", keep_default_na=True, skiprows=14)
    wvl = np.array(data.iloc[:, 0])
    spectrum = np.array(data.iloc[:, 1])
    return wvl, spectrum


def asymmetric_tukey_f(
    x: float, edge1: float, edge2: float, edge1_width: float, edge2_width: float
):
    if edge1_width < 0 or edge2_width < 0:
        edge1_width = 0
        edge2_width = 0
        print("Warning: negative tukey edge width; rectangular window computed")
    if abs(edge1_width) + abs(edge2_width) > 2 * abs(edge2 - edge1):
        raise Exception("Error: tukey edge width larger than admissible")
    if edge2 < edge1:
        e1 = edge2
        e2 = edge1
        w1 = edge2_width
        w2 = edge1_width
    elif edge1 < edge2:
        e1 = edge1
        e2 = edge2
        w1 = edge1_width
        w2 = edge2_width
    xmin = e1 - w1 / 2
    xmax = e2 + w2 / 2
    if x < xmin or x > xmax:
        y = 0
    elif xmin <= x and x < xmin + w1:
        y = (1 - np.cos((x - xmin) * np.pi / w1)) / 2
    elif xmin + w1 <= x and x <= xmax - w2:
        y = 1
    elif xmax - w2 < x and x <= xmax:
        y = (1 - np.cos((x - xmax) * np.pi / w2)) / 2
    else:
        print("x, xmax, xmin, w1, w2: ", x, xmax, xmin, w1, w2)
        raise ValueError(
            "in tukey_f, x could not be assigned correctly to the sub intervals, might it (or the other parameters) be NaN?"
        )
    return y


def asymmetric_tukey_window(
    x, edge1: float, edge2: float, edge1_width: float, edge2_width: float
):
    if isinstance(x, np.ndarray):
        y = []
        for xi in x:
            y.append(asymmetric_tukey_f(xi, edge1, edge2, edge1_width, edge2_width))
        y = np.array(y)
    elif isinstance(x, list):
        y = []
        for xi in x:
            y.append(asymmetric_tukey_f(xi, edge1, edge2, edge1_width, edge2_width))
    elif isinstance(x, float):
        y = asymmetric_tukey_f(x, edge1, edge2, edge1_width, edge2_width)
    else:
        raise TypeError(
            "in function tukey window x is neither np array nor list nor float"
        )
    return y


class SpectrumHandler:
    """Class to perform standard operations on spectra (calibration, plotting, division, multiplication, etc.)."""

    def __init__(
        self,
        filename: str = None,
        wavelengths=None,
        spectrum=None,
        remove_offsets_individually: bool = False,
        nm_smearing=1.0,
        eliminate_outliers_spectrum=False,
        filetype="MayaScarab",
    ):
        """Provide either a filename or the wavelengths and spectrum arrays to the constructor.

        Args:
            filename (str): Path to the file containing the spectrum data.
            wavelengths and spectrum (np.ndarray): = Wavelengths and spectrum data arrays. (alternative to filename)
            remove_offsets_individually (bool): = If True, compute and remove offsets from each column individually
                (filename is expected to have one column of wavelength data and arbitrarily many columns of equivalent spectra [to be averaged]).
            nm_smearing (float): Smearing in nm to be applied to the spectrum.
            eliminate_outliers_spectrum (bool): If True, eliminate single pixel outliers in the spectral data and replace them with average of surrounding values.
            filetype (str): Type of the file to be read. Options are 'MayaScarab' (default) or 'OceanOptics'.

        """
        self.nm_smearing = nm_smearing
        self.wvl, self.spectrum = np.linspace(200, 1000, 800), np.zeros(800)
        self.calibration_lamp_wvl, self.calibration_lamp_spectrum = None, None
        self.calibration_factor = None

        if filetype == "MayaScarab" and filename is not None:
            self.wvl, self.spectrum = read_spectrum_maya(
                filename,
                remove_offsets_individually,
                nm_smearing,
                eliminate_outliers_spectrum,
            )
        elif filetype == "OceanOptics" and filename is not None:
            self.wvl, self.spectrum = read_spectrum_ocean_optics(filename)
        elif wavelengths is not None and spectrum is not None:
            self.wvl = wavelengths
            self.spectrum = spectrum
        self.check_wvl_ascending()

    def get_spectrum(self):
        return self.wvl, self.spectrum

    def get_calibration_factor(self):
        if self.calibration_factor is None:
            print(
                "Warning: Calibration factor not available. Computing it from given spectra..."
            )
            self.compute_calibration_factor_spectrometer()
        return self.wvl, self.calibration_factor

    def calibrate(self):
        """Applies the calibration factor to the spectrum and sets the calibration factor to None."""
        if self.calibration_factor is None:
            print(
                "Warning: Calibration factor not available. Try load_calibration_factor_from_file(). Returning original spectrum."
            )
        else:
            self.spectrum = self.spectrum * self.calibration_factor
            self.calibration_factor = None
            print(
                "Calibration factor applied to the spectrum. Calibration factor is now set to None."
            )

    def save_to_file(self, filename):
        if self.wvl is None or self.spectrum is None:
            raise ValueError("Spectrum not loaded.")
        data = pd.DataFrame({"Wavelength": self.wvl, "Spectrum": self.spectrum})
        data.to_csv(filename, index=False, sep="\t")

    def save_calibration_factor_to_file(self, filename):
        """Saves the calibration factor to a file.

        The file can be either a .txt, .csv or .npz file.
        The file will contain the wavelength, calibration factor, calibration lamp true spectrum and measured lamp spectrum.
        """
        if self.calibration_factor is None:
            raise ValueError("Calibration factor not computed.")
        if ".txt" in filename or ".csv" in filename:
            data = pd.DataFrame(
                {
                    "Wavelength": self.wvl,
                    "Calibration Factor": self.calibration_factor,
                    "Lamp Spectrum": self.calibration_lamp_spectrum,
                    "Measured Spectrum": self.spectrum,
                }
            )
            data.to_csv(filename, index=False, sep="\t")
        elif ".npz" in filename:
            np.savez(
                filename,
                wavelength=self.wvl,
                corr_factor_smoothed=self.calibration_factor,
                lamp_ref=self.calibration_lamp_spectrum,
                lamp_measured=self.spectrum,
            )
        else:
            raise ValueError("File type not supported. Use .txt, .csv or .npz.")

    def load_calibration_factor_from_file(self, filename):
        data = pd.read_table(filename, sep="\t", keep_default_na=True, skiprows=0)
        self.calibration_factor = np.interp(
            self.wvl, np.array(data.iloc[:, 0]), np.array(data.iloc[:, 1])
        )

    def check_wvl_ascending(self):
        if self.wvl is not None and np.any(np.diff(self.wvl)) < 0:
            if np.all(np.diff(self.wvl) < 0):
                self.wvl = np.flip(self.wvl)
                self.spectrum = np.flip(self.spectrum)
            else:
                raise ValueError(
                    "Wavelength array must be either in ascending or descending order."
                )
        # same for calib spectrum
        if (
            self.calibration_lamp_wvl is not None
            and np.any(np.diff(self.calibration_lamp_wvl)) < 0
        ):
            if np.all(np.diff(self.calibration_lamp_wvl) < 0):
                self.calibration_lamp_wvl = np.flip(self.calibration_lamp_wvl)
                self.calibration_lamp_spectrum = np.flip(self.calibration_lamp_spectrum)
            else:
                raise ValueError(
                    "Wavelength array must be either in ascending or descending order."
                )
        return True

    def load_calibration_lamp_data(
        self, filename="./calibration_data/7315273LS-Deuterium-Halogen_CC-VISNIR.lmp"
    ):
        Data = pd.read_table(filename)
        self.calibration_lamp_wvl, self.calibration_lamp_spectrum = (
            np.array(Data.iloc[:, 0]),
            np.array(Data.iloc[:, 1]),
        )
        self.calibration_lamp_spectrum = gaussian_filter1d(
            self.calibration_lamp_spectrum,
            self.nm_smearing
            / (self.calibration_lamp_wvl[1] - self.calibration_lamp_wvl[0]),
        )
        self.check_wvl_ascending()

    def plot_calibration_data(
        self,
        low_lim=None,
        up_lim=None,
        low_lim_y=None,
        up_lim_y=None,
        wavelength_ROI: list = [420, 800],
    ):
        """Plot the calibration data of the lamp, the measured lamp spectrum, and the calibration factor.

        The method assumes that the true calibration lamp data and the measured spectrum of the calibration lamp are loaded in the object
        The normalization of the three curves is arbitrary. By default the region between 420 and 800 nm is used for normalization.

        Args:
            low_lim, up_lim(float): Wavelength limits for the x-axis.
            low_lim_y, up_lim_y (float): Intensity limits for the y-axis.
            wavelength_ROI (list): [wvl1, wvl2] = Wavelength range for the normalization of the calibration factor (for display purposes).
                Default is [420, 800] nm.

        """
        if low_lim is None or up_lim is None:
            low_lim = np.min(self.wvl)
            up_lim = np.max(self.wvl)
        if low_lim_y is None or up_lim_y is None:
            low_lim_y = 0
            up_lim_y = 3
        fig, ax = plt.subplots()
        ax.plot(
            self.wvl,
            self.spectrum / np.max(self.spectrum),
            label="spectrometer reading",
        )
        multipl_factor_lamp_plot = (
            self.spectrum[
                (self.wvl > wavelength_ROI[0]) & (self.wvl < wavelength_ROI[1])
            ].mean()
            / self.calibration_lamp_spectrum[
                (self.calibration_lamp_wvl > wavelength_ROI[0])
                & (self.calibration_lamp_wvl < wavelength_ROI[1])
            ].mean()
        )

        ax.plot(
            self.calibration_lamp_wvl,
            self.calibration_lamp_spectrum
            / np.max(self.spectrum)
            * multipl_factor_lamp_plot,
            label="actual lamp spectrum",
        )
        if self.calibration_factor is not None:
            ax.plot(self.wvl, self.calibration_factor, label="calibration factor")
        ax.set_xlabel("Wavelength (nm)")
        ax.set_ylabel("Intensity (Arb. unit)")
        ax.legend(loc="upper right")
        ax.set_xlim(low_lim, up_lim)
        ax.set_ylim(low_lim_y, up_lim_y)
        return fig

    def calibrate_wavelength_axis(self, intercept: float, slope: float):
        """Calibrate the wavelength axis of the spectrum using the given coefficient of a linear fit.

        Form of the calibration: true_wavelength = intercept + slope * measured_wavelength
        The function modifies the wavelength axis in place.

        Args:
            intercept: float,
            slope: float

        """
        if self.wvl is None or self.spectrum is None:
            raise ValueError("Spectrum not loaded.")
        self.wvl = intercept + slope * self.wvl
        self.check_wvl_ascending()

    def tukey_filter(
        self, edge1: float, edge2: float, edge1_width: float, edge2_width: float
    ):
        """Applies a Tukey window to the spectrum (e.g. to remove noise at the edges).

        Tukey window is defined as:
         * cosine-shaped window between edge1-edge1_width/2 and edge1+edge1_width/2
         * 1 between edge1+edge1_width/2 and edge2-edge2_width/2
         * cosine-shaped window between edge2-edge2_width/2 and edge2+edge2_width/2
         * 0 everywhere else
        In this case, tukey window is computed in the wavelength domain.

        Args:
            edge1, edge2 (float): Wavelength limits for the Tukey window (edge2-edge1 = FWHM, in nm)
            edge1_width, edge2_width (float): Width of the cosine-shaped edge of the tukey window (in nm).

        """
        window = asymmetric_tukey_window(
            self.wvl, edge1, edge2, edge1_width, edge2_width
        )
        self.spectrum = self.spectrum * window
        if self.calibration_factor is not None:
            self.calibration_factor = self.calibration_factor * window
        if (
            self.calibration_lamp_wvl is not None
            and self.calibration_lamp_spectrum is not None
        ):
            window = asymmetric_tukey_window(
                self.calibration_lamp_wvl, edge1, edge2, edge1_width, edge2_width
            )
            self.calibration_lamp_spectrum = self.calibration_lamp_spectrum * window

    def divide_by(self, spectrumObject, nm_smearing=0.0):
        """Normalizes the current spectrum with another spectrum, and stores the result in the present object.

        Args:
            spectrumObject (either 'c'/'calib'/'calibration'/'lamp' or SpectrumHandler object):
                the current spectrum is divided at each wavelength by the loaded calibration lamp data or by the spectrum passed as SpectrumHandler object.
            nm_smearing (float): sigma in nm for the gaussian smoothing

        """
        if (
            spectrumObject == "c"
            or spectrumObject == "calib"
            or spectrumObject == "calibration"
            or spectrumObject == "lamp"
            or spectrumObject == "calibration data"
        ):
            if (
                self.calibration_lamp_wvl is None
                or self.calibration_lamp_spectrum is None
            ):
                raise ValueError("Calibration lamp data not loaded.")
            wvl_divisor = self.calibration_lamp_wvl
            spectrum_divisor = self.calibration_lamp_spectrum
        elif not isinstance(spectrumObject, SpectrumHandler):
            raise TypeError("Input of divide_by must be a SpectrumHandler object.")
        else:
            wvl_divisor, spectrum_divisor = spectrumObject.get_spectrum()

        if nm_smearing > 0:
            spectrum_divisor = gaussian_filter1d(
                spectrum_divisor, sigma=nm_smearing / (wvl_divisor[1] - wvl_divisor[0])
            )
            self.spectrum = gaussian_filter1d(
                self.spectrum, sigma=nm_smearing / (self.wvl[1] - self.wvl[0])
            )
        commonWvl = self.wvl[
            (self.wvl > np.min(wvl_divisor)) & (self.wvl < np.max(wvl_divisor))
        ]
        interpolated_spd = np.interp(commonWvl, wvl_divisor, spectrum_divisor)
        interpolated_original = np.interp(commonWvl, self.wvl, self.spectrum)
        self.wvl = commonWvl
        self.spectrum = interpolated_original / interpolated_spd

    def add(self, spectrumObject):
        """Adds two spectra and stores the result in the present object.

        Args:
            spectrumObject (SpectrumHandler): spectrum to be added

        """
        if not isinstance(spectrumObject, SpectrumHandler):
            raise TypeError("Input of add must be a SpectrumHandler object.")
        wvl_add, spectrum_add = spectrumObject.get_spectrum()
        commonWvl = self.wvl[
            (self.wvl > np.min(wvl_add)) & (self.wvl < np.max(wvl_add))
        ]
        interpolated_spd = np.interp(commonWvl, wvl_add, spectrum_add)
        interpolated_original = np.interp(commonWvl, self.wvl, self.spectrum)
        self.wvl = commonWvl
        self.spectrum = interpolated_original + interpolated_spd

    def multiply(self, spectrumObject):
        """Multiplies the current spectrum by a second one at each wavelength.

        Args:
            spectrumObject (SpectrumHandler)

        """
        if not isinstance(spectrumObject, SpectrumHandler):
            raise TypeError("Input of multiply must be a SpectrumHandler object.")
        wvl_mult, spectrum_mult = spectrumObject.get_spectrum()
        commonWvl = self.wvl[
            (self.wvl > np.min(wvl_mult)) & (self.wvl < np.max(wvl_mult))
        ]
        interpolated_spd = np.interp(commonWvl, wvl_mult, spectrum_mult)
        interpolated_original = np.interp(commonWvl, self.wvl, self.spectrum)
        self.wvl = commonWvl
        self.spectrum = interpolated_original * interpolated_spd

    def multiply_scalar(self, scalar):
        """Multiplies the current spectrum by a constant.

        Args:
            scalar (float)

        """
        self.spectrum = self.spectrum * scalar

    def add_scalar(self, scalar):
        """Adds a constant offset.

        Args:
            scalar (float)

        """
        self.spectrum = self.spectrum + scalar

    def compute_calibration_factor_spectrometer(
        self,
        transmission_additional_optics=None,
        smoothing="poly",
        extend_calibration: bool = False,
        wavelength_ROI: list = [420, 800],
    ):
        """Computes the calibration factor.

        It assumes that in the SpectrumHandler object:
        * the measured spectrum of the calibration lamp is loaded as main spectrum
        * the true tabulated spectrum of the calibration lamp is loaded as self.calibration_lamp_spectrum [via load_calibration_lamp_data()]

        Args:
           transmission_additional_optics: list of SpectrumHandler objects = list of transmission spectra of additional optics to be taken into account when computing the calibration factor.
               The transmission ['reflection' for mirrors] spectra of additional optics will be included (multiplied) in the calibration factor
               Additional optics are assumed to be in the beam path of any standard measurement (e.g. integrating sphere) BUT NOT PRESENT during the calibration lamp measurement.
           smoothing: Type of smoothing to be applied to the calibration factor. Options are 'poly' (default) None or an integer number. If 'poly', a polynomial fit is applied to the calibration factor.
               If None, no smoothing is applied. If an integer number, a box filter of that size is applied to the calibration factor.
           extend_calibration (bool): If True, the calibration factor is extended to the full wavelength range of the spectrometer (the first and last values of the calibration curve are used for the extension).
               This is a quick and dirty solution to deal with the fact that the calibration lamp spectrum is not available for the full wavelength range of the spectrometer.
           wavelength_ROI (list): [wvl1, wvl2] Wavelength range for the normalization of the calibration factor (for display purposes).
               Default is [420, 800] nm.

        """
        if self.calibration_lamp_wvl is None or self.calibration_lamp_spectrum is None:
            raise ValueError("Calibration lamp data not loaded.")
        wvl_stored, spectrum_stored = self.wvl, self.spectrum
        self.divide_by("calibration data")

        eliminate_outliers(self.spectrum, threshold=3, window_points=20)

        if transmission_additional_optics is not None:
            for sao in transmission_additional_optics:
                if not isinstance(sao, SpectrumHandler):
                    raise TypeError(
                        "Input of spectra_additional_optics must be a list of SpectrumHandler objects."
                    )
                wvl_sao, spectrum_sao = sao.get_spectrum()
                if np.any(np.diff(wvl_sao)) < 0:
                    raise ValueError(
                        "Wavelength array of additional optics spectrum must be sorted in ascending order."
                    )
                wvl_sao = np.concatenate(
                    (
                        np.array([0]),
                        wvl_sao,
                        np.array([max(np.max(self.wvl), np.max(wvl_sao))]),
                    )
                )
                spectrum_sao = np.concatenate(
                    (
                        np.array([spectrum_sao[0]]),
                        spectrum_sao,
                        np.array([spectrum_sao[-1]]),
                    )
                )
                self.spectrum = self.spectrum * np.interp(
                    self.wvl, wvl_sao, spectrum_sao
                )

        self.calibration_factor = self.spectrum
        if smoothing is None:
            pass
        elif smoothing == "poly":
            smoothed_cf = np.poly1d(np.polyfit(self.wvl, self.calibration_factor, 50))
            self.calibration_factor = smoothed_cf(self.wvl)
        else:
            if not isinstance(smoothing, int):
                raise TypeError(
                    "Input of smoothing must be 'poly', None or an integer."
                )
            if smoothing < 1:
                raise ValueError("Input of smoothing must be a positive integer.")
            self.calibration_factor = box_smooth(self.calibration_factor, smoothing)

        for i in range(len(self.wvl)):
            if self.calibration_factor[i] < 0.01:
                self.calibration_factor[i] = 0.01
        self.calibration_factor = 1 / self.calibration_factor
        self.calibration_factor = self.calibration_factor / np.mean(
            self.calibration_factor[
                (self.wvl > wavelength_ROI[0]) & (self.wvl < wavelength_ROI[1])
            ]
        )

        if extend_calibration:
            # Extend the calibration factor to the full wavelength range of the spectrometer (pick the first and last values of the calibration curve for the extension)
            self.wvl = np.concatenate(
                (
                    np.linspace(0, self.wvl[0], 100),
                    self.wvl,
                    np.linspace(self.wvl[-1], self.wvl[-1], 100),
                )
            )
            self.calibration_factor = np.concatenate(
                (
                    np.ones(100) * self.calibration_factor[0],
                    self.calibration_factor,
                    np.ones(100) * self.calibration_factor[-1],
                )
            )
            self.calibration_factor = np.interp(
                wvl_stored, self.wvl, self.calibration_factor
            )
            self.wvl = wvl_stored
            self.spectrum = spectrum_stored
        else:
            self.spectrum = spectrum_stored[
                (wvl_stored > np.min(self.calibration_lamp_wvl))
                & (wvl_stored < np.max(self.calibration_lamp_wvl))
            ]

    def plot_spectrum(self, low_lim=None, up_lim=None):
        fig, ax = plt.subplots()
        if low_lim is not None and up_lim is not None:
            ax.plot(
                self.wvl[(self.wvl > low_lim) & (self.wvl < up_lim)],
                self.spectrum[(self.wvl > low_lim) & (self.wvl < up_lim)],
            )
        else:
            ax.plot(self.wvl, self.spectrum)
        ax.set_xlabel("Wavelength (nm)")
        ax.set_ylabel("Intensity (Arb. unit)")
        return fig


class MultiSpectrumHandler:
    """Stores and plots multiple spectra."""

    def __init__(
        self,
        filenameList: list = None,
        wavelengthList: list = None,
        spectrumList: list = None,
        spectrumHandlerList: list = None,
        remove_offsets_individually: bool = False,
        nm_smearing=1.0,
        eliminate_outliers_spectrum=False,
        filetype="MayaScarab",
    ):
        """Constructor for the MultiSpectrumHandler class.

        Args:
            filenameList: list = List of filenames to be read.
            wavelengthList: list = List of wavelength arrays. (alternative to filename)
            spectrumList: list = List of spectrum arrays. (alternative to filename)
            remove_offsets_individually: bool = If True, compute and remove offsets from each column individually
                (filename is expected to have one column of wavelength data and arbitrarily many columns of equivalent spectra [to be averaged]).
            nm_smearing: float = Smearing in nm to be applied to the spectrum.
            eliminate_outliers_spectrum: bool = If True, eliminate single pixel outliers in the spectral data and replace them with average of surrounding values.
            filetype: str = Type of the file to be read. Options are 'MayaScarab' (default) or 'OceanOptics'.

        """
        self.spectrumHandlers = []
        if filenameList is not None:
            for filename in filenameList:
                self.spectrumHandlers.append(
                    SpectrumHandler(
                        filename=filename,
                        remove_offsets_individually=remove_offsets_individually,
                        nm_smearing=nm_smearing,
                        eliminate_outliers_spectrum=eliminate_outliers_spectrum,
                        filetype=filetype,
                    )
                )
        elif wavelengthList is not None and spectrumList is not None:
            for i in range(len(wavelengthList)):
                self.spectrumHandlers.append(
                    SpectrumHandler(
                        wavelengths=wavelengthList[i],
                        spectrum=spectrumList[i],
                        remove_offsets_individually=remove_offsets_individually,
                        nm_smearing=nm_smearing,
                        eliminate_outliers_spectrum=eliminate_outliers_spectrum,
                        filetype=filetype,
                    )
                )
        elif spectrumHandlerList is not None:
            for spectrumHandler in spectrumHandlerList:
                if not isinstance(spectrumHandler, SpectrumHandler):
                    raise TypeError(
                        "Input of spectrumHandlerList must be a list of SpectrumHandler objects."
                    )
                self.spectrumHandlers.append(spectrumHandler)
        else:
            raise ValueError(
                "Either filenameList or wavelengthList and spectrumList or spectrumHandlerList must be provided. filetype, if provided, can only be 'MayaScarab' or 'OceanOptics'."
            )

    def plot(self, low_lim=None, up_lim=None):
        fig, ax = plt.subplots()
        for i, spectrumHandler in enumerate(self.spectrumHandlers):
            wvl, spectrum = spectrumHandler.get_spectrum()
            if low_lim is not None and up_lim is not None:
                ax.plot(
                    wvl[(wvl > low_lim) & (wvl < up_lim)],
                    spectrum[(wvl > low_lim) & (wvl < up_lim)],
                    label=f"Spectrum {i + 1}",
                )
            else:
                ax.plot(wvl, spectrum, label=f"Spectrum {i + 1}")
        ax.set_xlabel("Wavelength (nm)")
        ax.set_ylabel("Intensity (Arb. unit)")
        ax.legend(loc="upper right")
        return fig



# /// script
# [tool.marimo.display]
# theme = "dark"
# ///

import marimo

__generated_with = "0.14.13"
app = marimo.App(width="full")


@app.cell
def _(mo):
    mo.md(
        r"""
    ## Spectrometer calibration tool

    First, select the calibration lamp as measured by the spectrometer you want to calibrate.
    """
    )
    return


@app.cell
def _(mo):
    measurement_browser = mo.ui.file_browser(multiple=False)
    measurement_browser
    return (measurement_browser,)


@app.cell
def _(mo):
    mo.md(r"""Next, select the calibration lamp reference file from the list:""")
    return


@app.cell
def _(aw, mo):
    lamp_options = mo.ui.dropdown(
        options=[e.value for e in aw.spectrum.CalibrationLampReferences]
    )
    lamp_options
    return (lamp_options,)


@app.cell
def _(aw, measurement_browser, mo):
    _path = measurement_browser.path()
    mo.stop(_path is None)
    measurement = aw.data.load_mean_spectrum_from_scarab(_path).to_normalized()
    return (measurement,)


@app.cell
def _(aw, lamp_options, mo):
    mo.stop(lamp_options.value is None)
    reference = aw.data.load_mean_spectrum_from_scarab(
        aw.spectrum.get_calibration_path() / lamp_options.value
    ).to_normalized()
    return (reference,)


@app.cell
def _():
    return


@app.cell
def _(mo):
    mo.md(r"""Now, manually adjust the wavelength axis (optional - if it's not necessary, set offset and slope to zero)""")
    return


@app.cell
def _(mo):
    wavelength0 = mo.ui.slider(
        value=0.75,
        start=0,
        stop=3,
        step=1e-4,
        label="Wavelength shift center",
        show_value=True,
    )
    wavelength_offset = mo.ui.slider(
        value=-0.0107,
        start=-1.0,
        stop=1.0,
        step=1e-4,
        label="Wavelength offset",
        show_value=True,
    )
    wavelength_slope = mo.ui.slider(
        value=2e-3,
        start=-1,
        stop=1,
        step=1e-3,
        label="Wavelength slope",
        show_value=True,
    )
    mo.output.append(wavelength0)
    mo.output.append(wavelength_offset)
    mo.output.append(wavelength_slope)
    return wavelength0, wavelength_offset, wavelength_slope


@app.cell
def _(mo):
    mo.md(
        r"""
    Next, set the wavelength region-of-interest for the fitting, and the initial parameters of the response model. It is a slanted supergaussian. Minimize the difference between the calibrated measurement and the reference to give the fitting routine a good initial guess.

    Once it's decent, click run fitting.

    The "noise level" parameter will affect how the weights are adjusted based on the residuals of this fitting. If the noise level is zero, the weights will be set such that the calibrated spectrum exactly matches the reference, even for values where the measurement signal was very low. If the noise level is increased, the correction to the weights will be reduced for low signal-to-noise wavelengths.
    """
    )
    return


@app.cell
def _(measurement, mo, np):
    roi_lowest = mo.ui.number(
        value=np.min(measurement.wavelength_nm()),
        label="Shortest fitted wavelength (nm)",
    )
    roi_highest = mo.ui.number(
        value=np.max(measurement.wavelength_nm()),
        label="Longest fitted wavelength (nm)",
    )
    mo.output.append(roi_lowest)
    mo.output.append(roi_highest)

    return roi_highest, roi_lowest


@app.cell
def _(mo):
    amplitude_lam0 = mo.ui.slider(
        value=0.65,
        start=0,
        stop=3,
        step=1e-4,
        label="Amplitude correction center",
        show_value=True,
    )
    amplitude_offset = mo.ui.slider(
        value=3.0,
        start=0,
        stop=30,
        step=1e-4,
        label="Amplitude multiplier",
        show_value=True,
    )
    amplitude_slope = mo.ui.slider(
        value=4.0,
        start=-30,
        stop=30,
        step=1e-4,
        label="Amplitude slope",
        show_value=True,
    )
    amplitude_width = mo.ui.slider(
        value=0.32, start=0, stop=3, step=1e-4, label="Amplitude width", show_value=True
    )
    amplitude_order = mo.ui.slider(
        value=4.0,
        start=2,
        stop=32,
        step=1e-4,
        label="Amplitude Gaussian order",
        show_value=True,
    )
    wiener_noise_level = mo.ui.slider(
        value=0.01, start=0.0, stop=0.1, step=1e-4, label="Noise level", show_value=True
    )
    fit_button = mo.ui.run_button(label="Run fitting")
    fit_cb = mo.ui.checkbox(label="Fit continuously")
    mo.output.append(amplitude_lam0)
    mo.output.append(amplitude_offset)
    mo.output.append(amplitude_slope)
    mo.output.append(amplitude_width)
    mo.output.append(amplitude_order)
    mo.output.append(wiener_noise_level)
    mo.output.append(fit_button)
    mo.output.append(fit_cb)
    return (
        amplitude_lam0,
        amplitude_offset,
        amplitude_order,
        amplitude_slope,
        amplitude_width,
        fit_button,
        fit_cb,
        wiener_noise_level,
    )


@app.cell
def _(
    amplitude_lam0,
    amplitude_offset,
    amplitude_order,
    amplitude_slope,
    amplitude_width,
    aw,
    roi_highest,
    roi_lowest,
    wavelength0,
    wavelength_offset,
    wavelength_slope,
    wiener_noise_level,
):
    input_parameters = aw.data.CalibrationInput(
        wavelength_center=wavelength0.value,
        wavelength_offset=wavelength_offset.value,
        wavelength_slope=wavelength_slope.value,
        amplitude_center=amplitude_lam0.value,
        amplitude_multiplier=amplitude_offset.value,
        amplitude_slope=amplitude_slope.value,
        amplitude_width=amplitude_width.value,
        amplitude_order=amplitude_order.value,
        noise_level=wiener_noise_level.value,
        roi_lowest=1e-9 * roi_lowest.value,
        roi_highest=1e-9 * roi_highest.value,
    )
    return (input_parameters,)


@app.cell
def _(
    aw,
    fit_button,
    fit_cb,
    input_parameters,
    measurement,
    plot_xmax,
    plot_xmin,
    reference,
):
    if fit_button.value or fit_cb.value:
        calibration_dataset = aw.data.CalibrationDataset.generate(
            measurement=measurement,
            reference=reference,
            input_parameters=input_parameters,
        )
        calibration_dataset.plot(plot_xmax=plot_xmax.value, plot_xmin=plot_xmin.value)
    else:
        input_parameters.plot(
            measurement=measurement,
            reference=reference,
            plot_xmax=plot_xmax.value,
            plot_xmin=plot_xmin.value,
        )
    aw.plot.showmo()
    return (calibration_dataset,)


@app.cell
def _(measurement, mo, np):
    plot_xmin = mo.ui.number(
        value=np.min(measurement.wavelength_nm()), label="Plot min wavelength (nm)"
    )
    plot_xmax = mo.ui.number(
        value=np.max(measurement.wavelength_nm()), label="Plot max wavelength (nm)"
    )
    mo.output.append(plot_xmin)
    mo.output.append(plot_xmax)
    return plot_xmax, plot_xmin


@app.cell
def _(mo):
    mo.md(r"""If the result is good, save it as an .npz file, and contribute it to the database :)""")
    return


@app.cell
def _(mo):
    save_button = mo.ui.run_button(label="Save calibration")
    save_dataset_button = mo.ui.run_button(label="Save dataset")
    mo.output.append(save_button)
    mo.output.append(save_dataset_button)
    return save_button, save_dataset_button


@app.cell
def _(calibration_dataset, filedialog, mo, save_button):
    mo.stop(not save_button.value)

    _file_path = filedialog.asksaveasfilename(
        title="Save File", filetypes=[("npz files", "*.npz")]
    )

    if _file_path is not None:
        try:
            calibration_dataset.final_calibration.save_npz(_file_path)
        except NameError:
            print("Can't save without data")

    return


@app.cell
def _(calibration_dataset, filedialog, mo, save_dataset_button):
    mo.stop(not save_dataset_button.value)

    _file_path = filedialog.asksaveasfilename(
        title="Save File", filetypes=[("YAML files", "*.yml")]
    )

    if _file_path is not None:
        try:
            calibration_dataset.save_yaml(_file_path)
        except NameError:
            print("Can't save without data")
    return


@app.cell
def _():
    import attoworld as aw
    import marimo as mo

    aw.plot.set_style("nick_dark")
    import tkinter as tk
    from tkinter import filedialog

    import numpy as np

    root = tk.Tk()
    root.withdraw()
    return aw, filedialog, mo, np


if __name__ == "__main__":
    app.run()



"""Data class for handing complex spectra."""

import copy
from dataclasses import dataclass

import numpy as np

from .decorators import yaml_io


@yaml_io
@dataclass(slots=True)
class ComplexSpectrum:
    """Contains a complex spectrum, with spectral weights on a frequency scale.

    Attributes:
        spectrum (np.ndarray): the spectrum in complex128 format
        freq (np.ndarray):

    """

    spectrum: np.ndarray
    freq: np.ndarray

    def lock(self):
        """Make the data immutable."""
        self.spectrum.setflags(write=False)
        self.freq.setflags(write=False)

    def copy(self):
        """Return a copy of the data."""
        return copy.deepcopy(self)

    def to_time_derivative(self):
        r"""Return a ComplexSpectrum corresponding to the time derivative (multiply by $i\omega$)."""
        d_dt = 1j * 2 * np.pi * self.freq * self.spectrum
        return ComplexSpectrum(spectrum=d_dt, freq=np.array(self.freq))

    def to_bandpassed(self, frequency: float, sigma: float, order: int = 4):
        r"""Return the complex spectrum after applying a supergaussian bandpass filter to the spectrum, of the form
        $e^{\frac{(f-f_0)^r}{2\sigma^r}}$
        where $f_0$ is the frequency argument, $\sigma$ is the sigma argument, and r is the order argument.

        Args:
            frequency: the central frequency (Hz) of the bandpass
            sigma: the width of the bandpass (Hz)
            order: the order of the supergaussian

        """
        new_spectrum = self.spectrum * np.exp(
            -((self.freq - frequency) ** order) / (2 * sigma**order)
        )
        return ComplexSpectrum(spectrum=new_spectrum, freq=np.array(self.freq))



"""Base class for containing spectral data."""

import copy
from dataclasses import dataclass
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from scipy import constants

from ..numeric import (
    derivative,
    interpolate,
)
from .decorators import yaml_io


@yaml_io
@dataclass(slots=True)
class IntensitySpectrum:
    """Contains an intensity spectrum - real valued. SI units.

    Attributes:
        spectrum (Optional[np.ndarray]): the spectral intensity
        phase (Optional[np.ndarray]): the spectral phase
        freq (Optional[np.ndarray]): the frequencies corresponding to the spectrum
        wavelength (Optional[np.ndarray]): the wavelengths
        is_frequency_scaled (bool): has the lamda^-2 Jakobian been applied to Fourier transformed data?

    """

    spectrum: np.ndarray
    phase: Optional[np.ndarray]
    freq: np.ndarray
    wavelength: np.ndarray
    is_frequency_scaled: bool = False

    def lock(self):
        """Make the data immutable."""
        self.spectrum.setflags(write=False)
        self.freq.setflags(write=False)
        self.wavelength.setflags(write=False)
        if self.phase is not None:
            self.phase.setflags(write=False)

    def copy(self):
        """Make a copy of the class."""
        return copy.deepcopy(self)

    def wavelength_nm(self):
        """Returns:
        Optional[np.ndarray]: the wavelengths in nm.

        """
        if self.wavelength is not None:
            return 1e9 * self.wavelength
        return None

    def wavelength_micron(self):
        """Returns:
        Optional[np.ndarray]: the wavelengths in microns.

        """
        if self.wavelength is not None:
            return 1e6 * self.wavelength
        return None

    @staticmethod
    def from_spectrometer_spectrum_nanometers(
        wavelengths_nanometers: np.ndarray, spectrum: np.ndarray
    ):
        """Generate an instance based on a spepctrum array and wavelength array in nm (typical spectrometer data).

        Args:
            wavelengths_nanometers: the wavelengths in nanometers
            spectrum: the spectral intensity

        """
        wavelengths = 1e-9 * wavelengths_nanometers
        freqs = constants.speed_of_light / wavelengths
        phase = np.zeros(freqs.shape, dtype=float)
        return IntensitySpectrum(
            spectrum=spectrum, wavelength=wavelengths, freq=freqs, phase=phase
        )

    def get_wavelength_spectrum(self):
        """Return a wavelength-scaled spectrum, independent of the state of the current instance.

        Returns:
            np.ndarray, np.ndarray: wavelength and spectrum

        """
        if self.is_frequency_scaled:
            if self.freq is not None and self.spectrum is not None:
                from ..spectrum import frequency_to_wavelength

                return frequency_to_wavelength(self.freq, self.spectrum)
            raise Exception("Missing data")
        if self.wavelength is not None and self.spectrum is not None:
            return self.wavelength, self.spectrum
        raise Exception("Missing data")

    def get_frequency_spectrum(self):
        """Return a frequency-scaled spectrum, independent of the state of the current instance.

        Returns:
            np.ndarray, np.ndarray: wavelength and spectrum

        """
        if self.is_frequency_scaled:
            if self.freq is not None and self.spectrum is not None:
                return self.freq, self.spectrum
            raise Exception("Missing data")
        if self.wavelength is not None and self.spectrum is not None:
            from ..spectrum import wavelength_to_frequency

            return wavelength_to_frequency(1e9 * self.wavelength, self.spectrum)
        raise Exception("Missing data")

    def to_normalized(self):
        """Returns a normalized version of the current instance."""
        normalized_spectrum = self.spectrum / np.max(self.spectrum)

        return IntensitySpectrum(
            spectrum=normalized_spectrum,
            phase=self.phase,
            freq=np.array(self.freq),
            wavelength=np.array(self.wavelength),
            is_frequency_scaled=self.is_frequency_scaled,
        )

    def to_interpolated_wavelength(self, new_wavelengths: np.ndarray):
        """Replace the wavelength axis and interpolate the rest of the data onto it."""
        new_spectrum = interpolate(new_wavelengths, self.wavelength, self.spectrum)
        if self.phase is not None:
            new_phase = interpolate(new_wavelengths, self.wavelength, self.phase)
        else:
            new_phase = None

        return IntensitySpectrum(
            spectrum=new_spectrum,
            wavelength=new_wavelengths,
            freq=constants.speed_of_light / new_wavelengths,
            phase=new_phase,
            is_frequency_scaled=self.is_frequency_scaled,
        )

    def to_corrected_wavelength(self, new_wavelengths: np.ndarray):
        """Correct the wavelength axis by replacing it with new wavelengths."""
        assert len(new_wavelengths) == len(self.wavelength)
        return IntensitySpectrum(
            spectrum=self.spectrum,
            wavelength=new_wavelengths,
            freq=constants.speed_of_light / new_wavelengths,
            phase=self.phase,
            is_frequency_scaled=self.is_frequency_scaled,
        )

    def plot_with_group_delay(
        self,
        ax: Optional[Axes] = None,
        phase_blanking: float = 0.05,
        shift_from_centered=True,
        xlim=None,
    ):
        """Plot the spectrum and group delay curve.

        Args:
            ax: optionally plot onto a pre-existing matplotlib Axes
            phase_blanking: only show phase information (group delay) above this level relative to max intensity
            shift_from_centered (bool): if the pulse is centered, the group delay will be near +/- the grid length, tells whether to fix this.
            xlim: pass arguments to set_xlim() to constrain the x-axis

        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()

        start_index = np.argmax(self.spectrum > 0)
        intensity = self.spectrum[start_index::]
        freq = self.freq[start_index::]
        wl = constants.speed_of_light / freq
        if self.phase is not None:
            if shift_from_centered:
                shift = -0.5 / (freq[1] - freq[0])
                phase_shift = np.angle(np.exp(1j * 2 * np.pi * freq * shift))
                phase = np.unwrap(phase_shift + self.phase[start_index::])
            else:
                phase = np.unwrap(self.phase[start_index::])

        else:
            phase = np.zeros(intensity.shape, dtype=float)

        intensity /= np.max(intensity)
        intensity_line = ax.plot(1e9 * wl, intensity, label="Intensity")
        ax.set_xlabel("Wavelength (nm)")
        ax.set_ylabel("Intensity (Arb. unit)")
        ax_phase = plt.twinx(ax)
        group_delay = (-1e15 / (2 * np.pi)) * derivative(phase, 1) / (freq[1] - freq[2])
        assert isinstance(ax_phase, Axes)
        ax_phase.plot([], [])
        if self.phase is not None:
            phase_line = ax_phase.plot(
                1e9 * wl[intensity > phase_blanking],
                group_delay[intensity > phase_blanking],
                "--",
                label="Group delay",
            )
        else:
            phase_line = None
        ax_phase.set_ylabel("Group delay (fs)")
        if xlim is not None:
            ax.set_xlim(xlim)
            ax_phase.set_xlim(xlim)
        if phase_line is not None:
            lines = intensity_line + phase_line
        else:
            lines = intensity_line
        ax.legend(lines, [str(line.get_label()) for line in lines])
        return fig



from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.optimize
import scipy.signal
import scipy.special
from scipy import constants

from ...numeric import find_maximum_location, fwhm


def check_equal_length(*arg):
    n = len(arg[0])
    for v in arg:
        if len(v) != n:
            print(v)
            print("Error: vector size mismatch")
            raise Exception("Vector size mismatch")


def fourier_transform(TimeV, FieldV):  # complex!!!
    freq = np.fft.fftfreq(TimeV.size, d=TimeV[1] - TimeV[0])
    fft = np.fft.fft(FieldV)
    return freq, fft


def inverse_fourier_transform(freq, fullSpectrum):  # complex!!!
    timeV = np.fft.fftfreq(len(freq), freq[1] - freq[0])
    if len(timeV) % 2 == 0:
        timeV = np.concatenate(
            (timeV[int(len(timeV) / 2) :], timeV[: int(len(timeV) / 2)])
        )
    else:
        timeV = np.concatenate(
            (timeV[int((len(timeV) + 1) / 2) :], timeV[: int((len(timeV) + 1) / 2)])
        )
    fieldV = np.fft.ifft(fullSpectrum)

    return timeV, fieldV


def zero_padding(signalTimeV, signalV, fsExtension=50):
    timestep = signalTimeV[1] - signalTimeV[0]
    nPointsExtension = int(fsExtension / timestep)
    fsExtension = nPointsExtension * timestep
    newTimeV = np.linspace(
        signalTimeV[0] - fsExtension, signalTimeV[0] - timestep, nPointsExtension
    )
    newTimeV = np.append(newTimeV, signalTimeV[:])
    newTimeV = np.append(
        newTimeV,
        np.linspace(
            signalTimeV[-1] + timestep, signalTimeV[-1] + fsExtension, nPointsExtension
        )[:],
    )

    newSignalV = np.zeros(nPointsExtension)
    newSignalV = np.append(newSignalV, signalV[:])
    newSignalV = np.append(newSignalV, np.zeros(nPointsExtension)[:])

    trace_length = signalTimeV[-1] - signalTimeV[0]
    trace_center = signalTimeV[0] + trace_length / 2
    newSignalV = newSignalV * np.exp(
        -((newTimeV - trace_center) ** 10) / (trace_length * 9 / 20) ** 10
    )

    return newTimeV, newSignalV


def asymmetric_tukey_f(
    x: float, edge1: float, edge2: float, edge1_width: float, edge2_width: float
):
    if edge1_width < 0 or edge2_width < 0:
        edge1_width = 0
        edge2_width = 0
        print("Warning: negative tukey edge width; rectangular window computed")
    if abs(edge1_width) + abs(edge2_width) > 2 * abs(edge2 - edge1):
        raise Exception("Error: tukey edge width larger than admissible")
    if edge2 <= edge1:
        e1 = edge2
        e2 = edge1
        w1 = edge2_width
        w2 = edge1_width
    elif edge1 < edge2:
        e1 = edge1
        e2 = edge2
        w1 = edge1_width
        w2 = edge2_width
    xmin = e1 - w1 / 2
    xmax = e2 + w2 / 2
    if x < xmin or x > xmax:
        y = 0
    elif xmin <= x and x < xmin + w1:
        y = (1 - np.cos((x - xmin) * np.pi / w1)) / 2
    elif xmin + w1 <= x and x <= xmax - w2:
        y = 1
    elif xmax - w2 < x and x <= xmax:
        y = (1 - np.cos((x - xmax) * np.pi / w2)) / 2
    else:
        print("x, xmax, xmin, w1, w2: ", x, xmax, xmin, w1, w2)
        raise ValueError(
            "in tukey_f, x could not be assigned correctly to the sub intervals, might it (or the other parameters) be NaN?"
        )
    return y


def asymmetric_tukey_window(
    x, edge1: float, edge2: float, edge1_width: float, edge2_width: float
):
    if isinstance(x, np.ndarray):
        y = []
        for xi in x:
            y.append(asymmetric_tukey_f(xi, edge1, edge2, edge1_width, edge2_width))
        y = np.array(y)
    elif isinstance(x, list):
        y = []
        for xi in x:
            y.append(asymmetric_tukey_f(xi, edge1, edge2, edge1_width, edge2_width))
    elif isinstance(x, float):
        y = asymmetric_tukey_f(x, edge1, edge2, edge1_width, edge2_width)
    else:
        raise TypeError(
            "in function tukey window x is neither np array nor list nor float"
        )
    return y


class TraceHandler:
    """Loads and stores the trace (field vs time) saved to file (two columns, tab separated) or given in the form of time and field [or wavelengths, spectrum and phase] arrays.

    ALL THE TIMES ARE IN fs, FREQUENCIES IN PHz, WAVELENGTHS IN nm.

    Most of the class methods modify the object data in place, and don't return anything.

    Attributes:
        fieldTimeV: (WAVEFORM DATA in the time domain) time array (fs)
        fieldV: (WAVEFORM DATA in the time domain) field array
        fieldStdevV: (WAVEFORM DATA in the time domain) field standard deviation (or sigma) array

        frequencyAxis: (WAVEFORM DATA in the frequency domain) frequency axis of the FFT in PHz (should be np.fft compatible, i.e., equally spaced, in the form [0, df, 2*df, ..., fmax, -fmax[-df], ..., -df])
        fftFieldV: (WAVEFORM DATA in the frequency domain) FFT field (complex)
        complexFieldTimeV: (WAVEFORM DATA in the frequency domain) time axis of the complex field (fs)
        complexFieldV: (WAVEFORM DATA in the frequency domain) complex field ( = IFFT{FFT(ω)θ(ω)}(t), where θ(ω) = 1 if ω >= 0, θ(ω) = 0 otherwise )
        wvlAxis: (WAVEFORM DATA in the frequency domain) wavelength axis for the FFT spectrum (nm)
        fftSpectrum: (WAVEFORM DATA in the frequency domain) wavelength-dependent positive FFT spectrum ( = |FFT{field}|^2 * df/dλ, λ > 0 )
        fftphase: (WAVEFORM DATA in the frequency domain) wavelength-dependent spectral phase

        wvlSpectrometer: (SPECTROMETER DATA) wavelength (nm)
        ISpectrometer: (SPECTROMETER DATA) spectral intensity

        fsZeroPadding: (= 150 by default) time span in fs for zero padding (appending zeroes both before and after the trace, in order to get a smoother fft)
        filename: filename to load the trace from
        filename_spectrum: filename to load the spectrum from
        normalization_trace: normalization factor for the trace
        zero_delay: time zero, corresponds to the maximum envelope

    """

    def __init__(
        self,
        filename=None,
        filename_spectrum=None,
        time=None,
        field=None,
        stdev=None,
        wvl=None,
        spectrum=None,
        wvl_FFT_trace=None,
        spectrum_FFT_trace=None,
        phase_FFT_trace=None,
    ):
        """Constructor; loads the trace from file (or from time-field or wavelegth-spectrum-phase arrays) and stores it in the class.

        One of the following must be provided (as a default all args are None):
        * filename: the name of the file containing the trace (two columns, tab separated)
        * time and field: the time vector (fs) and electric field vector (a.u.)
        * wvl_FFT_trace, spectrum_FFT_trace and phase_FFT_trace: the wavelength (nm) spectrum (a.u.) and phase (rad) vectors

        Args:
            filename (str): the name of the file containing the trace (two columns - time and field, tab separated)
            filename_spectrum (str): the name of the file containing the spectrum (two columns - wavelength and spectral intensity, tab separated)
                this does not correspond to the fourier transform of the trace, but to some spectrometer data that you would like to compare to the loaded trace
            time (array): the time vector (fs)
            field (array): the electric field vector (a.u.)
            stdev (array): the standard deviation (or sigma) of the electric field vector (a.u.)
            wvl (array): the wavelength vector (nm) (spectrometer)
            spectrum (array): the spectral intensity vector (a.u.) (spectrometer)
            wvl_FFT_trace (array): the wavelength vector (nm) of the FFT trace (to reconstruct a trace from spectral data)
            spectrum_FFT_trace (array): the spectral intensity vector (a.u.) of the FFT trace (to reconstruct a trace from spectral data)
            phase_FFT_trace (array): the spectral phase vector (rad) of the FFT trace (to reconstruct a trace from spectral data)

        """
        self.fsZeroPadding = 150
        self.filename = filename
        self.filename_spectrum = filename_spectrum

        self.fieldTimeV = None
        self.fieldV = None
        self.fieldStdevV = None

        self.frequencyAxis = None
        self.fftFieldV = None
        self.complexFieldTimeV = None
        self.complexFieldV = None
        self.wvlAxis = None
        self.fftSpectrum = None
        self.fftphase = None

        self.wvlSpectrometer = None
        self.ISpectrometer = None

        self.normalization_trace = None
        self.zero_delay = None

        if filename is not None:
            self.load_trace()
        elif time is not None and field is not None:
            self.load_trace_from_arrays(time, field, stdev)
        elif (
            wvl_FFT_trace is not None
            and spectrum_FFT_trace is not None
            and phase_FFT_trace is not None
        ):
            self.load_trace_from_spectral_data(
                wvl_FFT_trace, spectrum_FFT_trace, phase_FFT_trace
            )
        else:
            print(
                "\n\nWARNING: in function TraceHandler.__init__() no data was loaded\n\n"
            )

        if self.filename_spectrum is not None:
            self.load_spectrum()
        elif wvl is not None and spectrum is not None:
            self.load_spectrum_from_arrays(wvl, spectrum)

    def load_spectrum_from_arrays(self, wvl, spectrum):
        """Loads the spectrum from wavelength and spectrum arrays and stores it in the class.

        The spectrum does not correspond to the fourier transform of the trace, but to some spectrometer data that you would like to compare with the loaded trace

        Args:
            wvl: wavelength array (nm)
            spectrum: spectral intensity array

        """
        """wavelength is in nm"""
        check_equal_length(wvl, spectrum)
        self.wvlSpectrometer = np.array(wvl)
        self.ISpectrometer = np.array(spectrum)

    def load_trace_from_arrays(self, fieldTimeV, fieldV, fieldStdevV=None):
        """Loads the trace from time and field arrays and stores it in the class.

        Args:
            fieldTimeV: time array (fs)
            fieldV: field array
            fieldStdevV: standard deviation - or sigma - to be associated to the trace (optional, default = None)

        """
        if fieldTimeV is None or fieldV is None:
            raise ValueError(
                "\nin function TraceHandler.load_from_arrays() too many arguments were None\n"
                "it might be that the constructor was erroneously used with no or too few arguments\n"
            )
        if np.max(np.abs(fieldTimeV)) < 1.0:
            print(
                "\n\nWARNING: do you remember that the time axis of TraceHandler is in fs?\n\n"
            )
        self.fieldTimeV = np.array(fieldTimeV)
        self.fieldV = np.array(fieldV)
        if fieldStdevV is not None:
            self.fieldStdevV = np.array(fieldStdevV)
        self.normalization_trace = np.max(np.abs(self.fieldV))
        self.update_fft()

    def load_trace(self, fname=None):
        """Loads from file."""
        if fname is not None:
            self.filename = fname
        data = pd.read_csv(self.filename, sep="\t")
        self.fieldTimeV = data["delay (fs)"].to_numpy()
        self.fieldV = data["field (a.u.)"].to_numpy()
        self.fieldStdevV = data["stdev field"].to_numpy()
        self.normalization_trace = np.max(np.abs(self.fieldV))
        self.update_fft()

    def load_trace_from_spectral_data(
        self, wvl_FFT_trace, spectrum_FFT_trace, phase_FFT_trace
    ):
        """Loads the trace from wavelength, spectrum, and spectral phase arrays and stores it in the class.

        This function has a different job than load_spectrum_from_arrays(): in fact, it retrieves the trace from the given spectral data via fourier transform.

        Args:
            wvl_FFT_trace: wavelength array in nm
            spectrum_FFT_trace: (wavelength-)spectral intensity array. This is assumed to be |FFT{field}|^2 * df/dλ
            phase_FFT_trace: spectral phase in rad

        """
        check_equal_length(wvl_FFT_trace, spectrum_FFT_trace, phase_FFT_trace)

        # check that the wvl_FFT_trace array is monotonically increasing
        if np.any(np.diff(wvl_FFT_trace) < 0):
            if np.all(np.diff(wvl_FFT_trace) < 0):
                wvl_FFT_trace = wvl_FFT_trace[::-1]
                spectrum_FFT_trace = spectrum_FFT_trace[::-1]
                phase_FFT_trace = phase_FFT_trace[::-1]
            else:
                raise ValueError(
                    "in function TraceHandler.load_from_spectral_data() wavelength array is not monotonous"
                )

        # frequency spectrum (monotonically increasing)
        freq = constants.speed_of_light / wvl_FFT_trace[::-1] * 1e-6
        spectrum_freq = spectrum_FFT_trace[::-1] / (
            constants.speed_of_light / wvl_FFT_trace[::-1] ** 2 * 1e-6
        )
        phase_FFT_trace = phase_FFT_trace[::-1]

        # create proper freq axis for the fourier transform
        df = (freq[1] - freq[0]) / 4
        nf = int(np.ceil(3 * freq[-1] / df))
        self.frequencyAxis = np.concatenate((np.arange(0, nf), np.arange(-nf, 0))) * df
        # corresponding time window
        twin = 1.0 / (df)

        # complex fourier transform
        spectrum_freq = np.sqrt(spectrum_freq) * np.exp(1.0j * phase_FFT_trace)

        # fill with zeros
        initial_zero_freq = np.linspace(
            freq[1] - freq[0], freq[0], int(np.ceil(4 * freq[0] / (freq[1] - freq[0])))
        )
        initial_zeros = np.zeros(len(initial_zero_freq))
        final_zero_freq = np.linspace(
            2 * freq[-1] - freq[-2],
            4 * freq[-1],
            int(np.ceil(15 * freq[-1] / (freq[-1] - freq[-2]))),
        )
        final_zeros = np.zeros(len(final_zero_freq))
        freq = np.concatenate((initial_zero_freq, freq, final_zero_freq))
        spectrum_freq = np.concatenate((initial_zeros, spectrum_freq, final_zeros))
        # add negative frequencies
        freq = np.concatenate((-freq[::-1], freq))
        spectrum_freq = np.concatenate(
            (np.conjugate(spectrum_freq[::-1]), spectrum_freq)
        )

        # interpolate the spectrum to the frequency axis of the fft
        self.fftFieldV = np.interp(self.frequencyAxis, freq, spectrum_freq)

        # add linear phase to center the pulse in the time window; this has to be done after interpolation because the spectral phase will be fast oscillating
        self.fftFieldV = self.fftFieldV * np.exp(
            1.0j * 2 * np.pi * self.frequencyAxis * twin / 2
        )

        self.update_fft_spectrum()
        self.update_trace_from_fft()

    def load_spectrum(self, fname=None):
        """Loads spectrum from file.

        The spectrum does not correspond to the fourier transform of the trace, but to the spectrometer data that you would like to compare with the loaded trace

        Args:
            fname: filename

        """
        if fname is not None:
            self.filename_spectrum = fname
        data = pd.read_csv(self.filename_spectrum, sep="\t")
        self.wvlSpectrometer = data["wavelength (nm)"].to_numpy()
        self.ISpectrometer = data["intensity (a.u.)"].to_numpy()
        self.normalize_spectrum()

    def update_fft(self, zero_pad_field=True):
        """Updates the fft of the trace from the time domain data.

        Args:
            zero_pad_field: bool; if true (default) add zeros before and after the trace before computing fft (for a smoother spectrum)
                The length of appended zeros is defined by self.fsZeroPadding.

        """
        zp_fieldTimeV, zp_fieldV = deepcopy(self.fieldTimeV), deepcopy(self.fieldV)
        if zero_pad_field:
            zp_fieldTimeV, zp_fieldV = zero_padding(
                self.fieldTimeV, self.fieldV, fsExtension=self.fsZeroPadding
            )
        self.frequencyAxis, self.fftFieldV = fourier_transform(zp_fieldTimeV, zp_fieldV)
        self.update_fft_spectrum()

    def update_fft_spectrum(self):
        """Updates the wavelength-dependent intensity spectrum. Results are stored in self.wvlAxis and self.fftSpectrum.

        This is (derived from but) different from the fourier transform because:
         * it considers the squared modulus of the field
         * it multiplies the frequency-dependent spectral density by df/dλ = c/λ²
         * it considers only positive wavelengths

        The function internally calls update_spectral_phase() as well.
        """
        self.wvlAxis = (
            constants.speed_of_light
            / self.frequencyAxis[self.frequencyAxis > 0]
            * 1.0e-6
        )
        self.fftSpectrum = (
            np.abs(self.fftFieldV[self.frequencyAxis > 0]) ** 2
            * constants.speed_of_light
            / self.wvlAxis**2
        )
        self.compute_complex_field()
        self.update_spectral_phase()
        self.normalize_fft_spectrum()

    def update_spectral_phase(self):
        """Computes the spectral phase of the trace. The result is stored in self.fftphase.

        An attempt of automatic linear phase subtraction is made. Phase is unwrapped.
        """
        # determine the time distance (delay) of the waveform peak from the beginning of the trace (more precisely, from the beginning of the ifft)
        tmax = self.get_zero_delay()
        deltaT1 = tmax - self.complexFieldTimeV[0]
        deltaT2 = self.complexFieldTimeV[-1] - tmax
        ifft_twindow = 1 / (self.frequencyAxis[1] - self.frequencyAxis[0])
        total_delay = (
            ifft_twindow / 2 + (deltaT2 - deltaT1) / 2
        )  # total estimated delay of the peak from the beginning of the ifft trace

        # compute the phase, use the computed 'delay' to correct for the non-interesting linear component, and unwrap it
        self.fftphase = np.angle(self.fftFieldV[self.frequencyAxis > 0])
        self.fftphase = (
            self.fftphase
            - total_delay * self.frequencyAxis[self.frequencyAxis > 0] * 2 * np.pi
        )
        self.fftphase = np.unwrap(self.fftphase)

        # subtract constant phase offset (m * 2*π)
        n_pi = np.sum(self.fftSpectrum * self.fftphase) / np.sum(self.fftSpectrum)
        n_pi = int(n_pi / (2 * np.pi))
        self.fftphase -= n_pi * 2 * np.pi

    def update_trace_from_fft(self):  # remember to use together with strip_from_trace
        # careful: the complex part of the IFFT is discarded
        """Updates the field trace from fft.

        In most cases, you might want to call strip_from_trace() afterward, to eliminate the zero-padding from the trace.
        """
        self.fieldTimeV, self.fieldV = inverse_fourier_transform(
            self.frequencyAxis, self.fftFieldV
        )
        self.fieldV = np.real(self.fieldV)
        self.normalization_trace = np.max(np.abs(self.fieldV))
        self.get_zero_delay()

    def strip_from_trace(self, timeRange=None):
        """Eliminates the zeros appended to the trace (zero-padding) when computing the FFT.

        Useful when retrieving the trace via ifft (method update_trace_from_fft()).

        Args:
            timeRange (float): the time range (fs) to strip from the beginning and end of the trace. If None (default), fsZeroPadding is used.

        """
        if timeRange is None:
            timeRange = self.fsZeroPadding
        # strip the first and last fsZeroPadding time points
        self.fieldV = self.fieldV[
            (self.fieldTimeV >= np.min(self.fieldTimeV) + timeRange)
            & (self.fieldTimeV <= np.max(self.fieldTimeV) - timeRange)
        ]
        self.fieldTimeV = self.fieldTimeV[
            (self.fieldTimeV >= np.min(self.fieldTimeV) + timeRange)
            & (self.fieldTimeV <= np.max(self.fieldTimeV) - timeRange)
        ]
        if self.fieldStdevV is None:
            return True
        if len(self.fieldStdevV) > len(self.fieldV):
            n_remove = len(self.fieldStdevV) - len(self.fieldV)
            print(
                "\n\nWarning: fieldStdevV is longer than fieldV, removing",
                n_remove,
                " elements\n\n",
            )
            while n_remove > 0:
                if n_remove % 2 == 0:
                    self.fieldStdevV = np.delete(self.fieldStdevV, 0)
                else:
                    self.fieldStdevV = np.delete(self.fieldStdevV, -1)
                n_remove -= 1
        elif len(self.fieldStdevV) < len(self.fieldV):
            n_add = len(self.fieldV) - len(self.fieldStdevV)
            print(
                "\n\nWarning: fieldStdevV is shorter than fieldV, adding",
                n_add,
                " elements\n\n",
            )
            while n_add > 0:
                if n_add % 2 == 0:
                    self.fieldStdevV = np.insert(self.fieldStdevV, 0, 0)
                else:
                    self.fieldStdevV = np.append(self.fieldStdevV, 0)
                n_add -= 1
        return True

    def strip_from_complex_trace(self, timeRange=None):
        """From the complex trace eliminates the zeros appended (zero-padding) when computing the FFT (Analogous to strip_from_trace())."""
        if timeRange is None:
            timeRange = self.fsZeroPadding
        # strip the first and last fsZeroPadding time points
        self.complexFieldV = self.complexFieldV[
            (self.complexFieldTimeV >= np.min(self.complexFieldTimeV) + timeRange)
            & (self.complexFieldTimeV <= np.max(self.complexFieldTimeV) - timeRange)
        ]
        self.complexFieldTimeV = self.complexFieldTimeV[
            (self.complexFieldTimeV >= np.min(self.complexFieldTimeV) + timeRange)
            & (self.complexFieldTimeV <= np.max(self.complexFieldTimeV) - timeRange)
        ]

    def tukey_time_window(self, lowEdge, upEdge, lowEdgeWidth, upEdgeWidth):
        """Applies a tukey window to the trace in the time domain.

        Args:
            lowEdge, upEdge: lower and upper edge of the window upEdge-lowEdge = FWHM of the window
            lowEdgeWidth, upEdgeWidth: width of the cosine-shaped edges (from 0 to 1 or viceversa)

        """
        window = asymmetric_tukey_window(
            np.abs(self.fieldTimeV), lowEdge, upEdge, lowEdgeWidth, upEdgeWidth
        )
        self.fieldV = self.fieldV * window
        self.update_fft()

    def fourier_interpolation(self, ntimes_finer: int):
        """Shrinks the time step of the trace in time domain by a factor ntimes_finer.

        This is done by 'zero-padding' the fourier transform of the trace, i.e., by extending the frequency axis by a factor ntimes_finer, and adding corresponding zeros to the FFT field.

        Args:
            ntimes_finer (int): the factor by which to shrink the time step of the trace in time domain. Must be >= 1.

        """
        if ntimes_finer < 1:
            raise ValueError(
                "in function TraceHandler.fourier_interpolation() ntimes_finer must be >= 1"
            )
        if self.frequencyAxis is None or self.fftFieldV is None:
            raise ValueError(
                "in function TraceHandler.fourier_interpolation() frequency axis or fft field is not defined"
            )
        # extend the frequency axis by a factor ntimes_finer
        df = self.frequencyAxis[1] - self.frequencyAxis[0]
        i_last_pos = int(np.ceil(len(self.frequencyAxis) / 2) - 1)
        if self.frequencyAxis[i_last_pos] < 0 or self.frequencyAxis[i_last_pos + 1] > 0:
            print(self.frequencyAxis[0 : i_last_pos + 1])
            print(self.frequencyAxis[i_last_pos + 1 :])
            raise Exception(
                "in function fourier_interpolation() frequency axes was not extended correctly"
            )

        appended_freqFFT = np.linspace(
            self.frequencyAxis[i_last_pos] + df,
            ntimes_finer * self.frequencyAxis[i_last_pos] + df,
            round((ntimes_finer - 1) * self.frequencyAxis[i_last_pos] / df),
        )
        prepended_freqFFT = np.linspace(
            -(ntimes_finer - 1) * self.frequencyAxis[i_last_pos]
            + self.frequencyAxis[i_last_pos + 1],
            +self.frequencyAxis[i_last_pos + 1],
            round((ntimes_finer - 1) * self.frequencyAxis[i_last_pos] / df),
        )
        self.frequencyAxis = np.concatenate(
            (
                self.frequencyAxis[: i_last_pos + 1],
                appended_freqFFT,
                prepended_freqFFT,
                self.frequencyAxis[i_last_pos + 1 :],
            )
        )
        self.fftFieldV = np.concatenate(
            (
                self.fftFieldV[: i_last_pos + 1],
                np.zeros(len(appended_freqFFT)),
                np.zeros(len(prepended_freqFFT)),
                self.fftFieldV[i_last_pos + 1 :],
            )
        )

        self.update_trace_from_fft()
        self.strip_from_trace()
        self.update_fft_spectrum()

    def differentiate_trace(self, spectrally=True):
        """Computes the time derivative of the trace.

        Args:
            spectrally: bool; if True (default) the derivative will be computed in the spectral domain (maybe more stable numerically)

        """
        if spectrally:
            self.fftFieldV = 1j * self.frequencyAxis * self.fftFieldV
            self.update_trace_from_fft()
            self.update_fft_spectrum()
        else:
            self.fieldV = np.gradient(self.fieldV, self.fieldTimeV)
            self.update_fft()

    def integrate_trace(self, spectrally=True):
        """Integrates the trace in time.

        Args:
            spectrally (bool): if True (default) the integral will be computed in the spectral domain (maybe more stable numerically)

        """
        if spectrally:
            if self.frequencyAxis[0] != 0:
                print(
                    "WARNING: in function TraceHandler.integrate_trace() the FFT axis' first element is not zero"
                )
            self.fftFieldV[1:] = -1j / self.frequencyAxis[1:] * self.fftFieldV[1:]
            self.update_trace_from_fft()
            self.update_fft_spectrum()
        else:
            self.fieldV = np.cumsum(self.fieldV) * np.mean(np.diff(self.fieldTimeV))
            self.update_fft()

    def save_trace_to_file(self, filename, low_lim=None, up_lim=None):
        """Saves the trace to file.

        Args:
            filename
            low_lim, up_lim: only save the trace between low_lim and up_lim (default: None, None)

        """
        if low_lim is not None and up_lim is not None:
            fieldTimeV_to_write = self.fieldTimeV[
                (self.fieldTimeV >= low_lim) & (self.fieldTimeV <= up_lim)
            ]
            fieldV_to_write = self.fieldV[
                (self.fieldTimeV >= low_lim) & (self.fieldTimeV <= up_lim)
            ]
        else:
            fieldTimeV_to_write = self.fieldTimeV
            fieldV_to_write = self.fieldV
        data = pd.DataFrame(
            {"delay (fs)": fieldTimeV_to_write, "field (a.u.)": fieldV_to_write}
        )
        data.to_csv(filename, sep="\t", index=False)

    def normalize_spectrum(self):
        """Normalizes the comparison spectrum to its integral."""
        spectrum_range = [60, 930]
        n_factor = (
            np.abs(
                np.diff(self.wvlSpectrometer)[
                    (self.wvlSpectrometer[:-1] > spectrum_range[0])
                    & (self.wvlSpectrometer[:-1] < spectrum_range[1])
                ]
            )
            * self.ISpectrometer[:-1][
                (self.wvlSpectrometer[:-1] > spectrum_range[0])
                & (self.wvlSpectrometer[:-1] < spectrum_range[1])
            ]
        ).sum()
        self.ISpectrometer /= n_factor

    def normalize_fft_spectrum(self, spectrum_range=[60, 930]):
        """Normalizes the spectrum of the trace (|FFT{trace}|^2 * df/dλ) to its integral (area).

        Args:
            spectrum_range (list): the range of wavelengths (nm) to consider for the normalization. Default is [60, 930] nm.

        """
        n_factor = (
            np.abs(
                np.diff(self.wvlAxis)[
                    (self.wvlAxis[:-1] > spectrum_range[0])
                    & (self.wvlAxis[:-1] < spectrum_range[1])
                ]
            )
            * self.fftSpectrum[:-1][
                (self.wvlAxis[:-1] > spectrum_range[0])
                & (self.wvlAxis[:-1] < spectrum_range[1])
            ]
        ).sum()
        self.fftSpectrum /= n_factor

    def get_trace(self):
        """Returns the time (fs) and field arrays. Field is in a.u. unless set_fluence() was called or a calibrated field was given as an input."""
        return self.fieldTimeV, self.fieldV

    def get_spectrum_trace(self):
        """Returns the wavelength and spectral intensity corresponding to the fourier transform of the trace."""
        return self.wvlAxis, self.fftSpectrum

    def get_spectral_phase(self):
        """Returns the wavelength array and the spectral phase array corresponding to the FFT of the trace."""
        return self.wvlAxis, self.fftphase

    def get_stdev(self):
        """Returns only the standard deviation (sigma) array."""
        return self.fieldStdevV

    def get_positive_fft_field(self):
        """Returns the positive frequency axis f and the corresponding FFT field (complex).

        Main task of this function is to 'smooth' the phase of the fft (= remove linear phase, corresponding to a time shift in the temporal domain), since the waveform is usually centered around t = 0
        """
        if self.frequencyAxis is None or self.fftFieldV is None:
            raise ValueError(
                "in function TraceHandler.get_positive_fft_field() frequency axis or fft field is not defined"
            )
        positive_freq = self.frequencyAxis[self.frequencyAxis >= 0]
        positive_fft_field = self.fftFieldV[self.frequencyAxis >= 0]
        # remove linear phase
        tmax = self.get_zero_delay()
        deltaT1 = tmax - self.fieldTimeV[0]
        deltaT2 = self.fieldTimeV[-1] - tmax
        ifft_twindow = 1 / (self.frequencyAxis[1] - self.frequencyAxis[0])
        total_delay = ifft_twindow / 2 + (deltaT2 - deltaT1) / 2
        positive_fft_field = positive_fft_field * np.exp(
            -1.0j * 2 * np.pi * positive_freq * total_delay
        )
        return positive_freq, positive_fft_field

    def get_fluence(self):
        """Calculates the fluence of the trace.

        convention: field [V/Å], time [fs], fluence [J/cm²]
        F = c eps_0 integral(E^2)dt


        Returns:
            fluence: float

        """
        return (
            constants.speed_of_light
            * constants.epsilon_0
            * np.trapz(self.fieldV**2, self.fieldTimeV)
            * 1e1
        )

    def set_fluence(self, fluence):
        """Set the fluence of the trace.

        Convention: field [V/Å], time [fs], fluence [J/cm²]

        F = c eps_0 integral(E^2)dt
        """
        F = (
            constants.speed_of_light
            * constants.epsilon_0
            * np.trapz(self.fieldV**2, self.fieldTimeV)
            * 1e1
        )
        self.fieldV *= np.sqrt(fluence / F)
        self.update_fft()
        self.update_fft_spectrum()

    def compute_complex_field(self):
        """Computes the complex trace by inverting the positive-freq FFT and stores it in self.complexFieldV (for envelope or phase computation, for example).

        Notice that the fourier transform of the trace is not computed, it is assumed to be already stored in the class. This should be true in most cases.
        """
        complex_spectrum = deepcopy(self.fftFieldV)
        complex_spectrum[self.frequencyAxis < 0] = 0
        self.complexFieldTimeV, self.complexFieldV = inverse_fourier_transform(
            self.frequencyAxis, complex_spectrum
        )
        self.complexFieldV = 2 * self.complexFieldV
        self.strip_from_complex_trace()

    def get_envelope(self):
        """Get the envelope of the trace.

        Returns:
            time array (fs)
            envelope array

        """
        if self.complexFieldV is None:
            self.compute_complex_field()
        return self.complexFieldTimeV, np.abs(self.complexFieldV)

    def get_phase(self):
        """Get the instantaneous phase of the trace.

        If complexFieldV is already computed and stored, no re-calculation occurs

        Returns:
            time array (fs)
            instantaneous phase array (fs)

        """
        if self.complexFieldV is None:
            self.compute_complex_field()
        return self.complexFieldTimeV, np.angle(self.complexFieldV)

    def get_zero_delay(self):
        """Get the time value corresponding to the envelope peak.

        Returns:
            zero_delay: float

        """
        # careful: the time grid should be fine enough to resolve the maximum of the envelope
        t, en = self.get_envelope()
        dt = t[1] - t[0]
        max_index, max_value = find_maximum_location(en)
        self.zero_delay = t[0] + dt * max_index
        return self.zero_delay

    def get_FWHM(self):
        """Get the FWHM of the trace.

        Returns:
            FWHM: float

        """
        t, en = self.get_envelope()
        dt = t[1] - t[0]
        return fwhm(en**2, dt)

    def fft_tukey_bandpass(
        self, lowWavelengthEdge, upWavelengthEdge, lowEdgeWidth, highEdgeWidth
    ):
        """Applies a bandpass filter to the trace in the frequency domain using a tukey window.

        The tukey window is 1 between  lowWavelengthEdge+lowEdgeWidth/2  and  upWavelengthEdge-highEdgeWidth/2
        and it is reaches zero at lowWavelengthEdge-lowEdgeWidth/2  and  upWavelengthEdge+highEdgeWidth/2.
        Notice that the edges are only cosine-shaped in the frequency domain. In the wavelength domain upWavelengthEdge - lowWavelengthEdge does not coincide with the FWHM of the tukey function

        Args:
            lowWavelengthEdge: float (nm)
            upWavelengthEdge: float (nm)
            lowEdgeWidth: float (nm)
            highEdgeWidth: float (nm)

        """
        upFreqEdge = (
            constants.speed_of_light / (lowWavelengthEdge - lowEdgeWidth / 2) / 2 * 1e-6
            + constants.speed_of_light
            / (lowWavelengthEdge + lowEdgeWidth / 2)
            / 2
            * 1e-6
        )
        lowFreqEdge = (
            constants.speed_of_light / (upWavelengthEdge - highEdgeWidth / 2) / 2 * 1e-6
            + constants.speed_of_light
            / (upWavelengthEdge + highEdgeWidth / 2)
            / 2
            * 1e-6
        )

        upFreqEdgeWidth = (
            constants.speed_of_light / (lowWavelengthEdge - lowEdgeWidth / 2) * 1e-6
            - constants.speed_of_light / (lowWavelengthEdge + lowEdgeWidth / 2) * 1e-6
        )
        lowFreqEdgeWidth = (
            constants.speed_of_light / (upWavelengthEdge - highEdgeWidth / 2) * 1e-6
            - constants.speed_of_light / (upWavelengthEdge + highEdgeWidth / 2) * 1e-6
        )

        window = asymmetric_tukey_window(
            np.abs(self.frequencyAxis),
            lowFreqEdge,
            upFreqEdge,
            lowFreqEdgeWidth,
            upFreqEdgeWidth,
        )
        self.fftFieldV = self.fftFieldV * window
        self.update_trace_from_fft()
        self.update_fft_spectrum()
        self.strip_from_trace()

    def apply_transmission(self, wavelengths, f):
        """Applies a spectral transmission function to the trace (e.g. spectral filter).

        Args:
            wavelengths: ndarray = wavelength array (nm)
            f: ndarray = transmission function f(λ)

        """
        if np.any(np.diff(wavelengths)) < 0:
            if np.all(np.diff(wavelengths)) < 0:
                wavelengths = wavelengths[::-1]
                f = f[::-1]
            else:
                raise ValueError(
                    "in function TraceHandler.apply_transmission() wavelength array is not monotonous"
                )
        freq = constants.speed_of_light / wavelengths[::-1] * 1e-6
        spectrum_freq = f[::-1]

        # fill with zeros
        initial_zero_freq = np.linspace(
            freq[1] - freq[0], freq[0], int(np.ceil(4 * freq[0] / (freq[1] - freq[0])))
        )
        initial_zeros = np.zeros(len(initial_zero_freq))
        final_zero_freq = np.linspace(
            2 * freq[-1] - freq[-2],
            4 * freq[-1],
            int(np.ceil(15 * freq[-1] / (freq[-1] - freq[-2]))),
        )
        final_zeros = np.zeros(len(final_zero_freq))
        freq = np.concatenate((initial_zero_freq, freq, final_zero_freq))
        spectrum_freq = np.concatenate((initial_zeros, spectrum_freq, final_zeros))
        # add negative frequencies
        freq = np.concatenate((-freq[::-1], freq))
        spectrum_freq = np.concatenate((spectrum_freq[::-1], spectrum_freq))

        # interpolate the spectrum to the frequency axis of the fft
        spectrum_interp = np.interp(self.frequencyAxis, freq, spectrum_freq)

        self.fftFieldV = self.fftFieldV * spectrum_interp
        self.update_trace_from_fft()
        self.update_fft_spectrum()
        self.strip_from_trace()

    def apply_spectrum(
        self,
        wvl=None,
        spectrum=None,
        CEP_shift: float = 0.0,
        stripZeroPadding: bool = True,
    ):
        """Applies a spectrum to the phase of the trace. This means that a new trace is computed and stored in the TraceHandler object (replacing the existing one);
        the new trace has spectral intensity equal to the applied spectrum and spectral phase equal to the phase of the existing trace.

        Args:
            wvl: wavelength array (nm). If None (default) the comparison spectrum stored in the class (self.wvlSpectrometer and self.ISpectrometer) is applied.
            spectrum: spectral intensity array. If None (default) the comparison spectrum stored in the class (self.wvlSpectrometer and self.ISpectrometer) is applied.
            CEP_shift: a possible artificial phase shift IN UNITS OF PI! (default = 0)
            stripZeroPadding (bool): whether to eliminate the zero padding (zeros appended to the trace). Default is True

        """
        if wvl is None or spectrum is None:
            wvl = self.wvlSpectrometer
            spectrum = self.ISpectrometer
        # check that the wvl array is monotonically increasing
        if np.any(np.diff(wvl) < 0):
            if np.all(np.diff(wvl) < 0):
                wvl = wvl[::-1]
                spectrum = spectrum[::-1]
            else:
                raise ValueError(
                    "in function TraceHandler.apply_spectrum() wavelength array is not monotonous"
                )

        # frequency spectrum (monotonically increasing)
        freq = constants.speed_of_light / wvl[::-1] * 1e-6
        spectrum_freq = spectrum[::-1] / (
            constants.speed_of_light / wvl[::-1] ** 2 * 1e-6
        )
        spectrum_freq = np.maximum(spectrum_freq, np.zeros(len(spectrum_freq)))

        # fill with zeros
        initial_zero_freq = np.linspace(
            freq[1] - freq[0], freq[0], int(np.ceil(4 * freq[0] / (freq[1] - freq[0])))
        )
        initial_zeros = np.zeros(len(initial_zero_freq))
        final_zero_freq = np.linspace(
            2 * freq[-1] - freq[-2],
            4 * freq[-1],
            int(np.ceil(15 * freq[-1] / (freq[-1] - freq[-2]))),
        )
        final_zeros = np.zeros(len(final_zero_freq))
        freq = np.concatenate((initial_zero_freq, freq, final_zero_freq))
        spectrum_freq = np.concatenate((initial_zeros, spectrum_freq, final_zeros))
        # add negative frequencies
        freq = np.concatenate((-freq[::-1], freq))
        spectrum_freq = np.concatenate((spectrum_freq[::-1], spectrum_freq))

        # interpolate the spectrum to the frequency axis of the fft
        spectrum_interp = np.interp(self.frequencyAxis, freq, spectrum_freq)
        self.fftFieldV = np.sqrt(spectrum_interp) * np.exp(
            1j * np.angle(self.fftFieldV)
            + 1j * np.pi * CEP_shift * np.sign(self.frequencyAxis)
        )

        self.update_trace_from_fft()
        if stripZeroPadding:
            self.strip_from_trace()
        else:
            self.fieldStdevV = None
        self.update_fft_spectrum()

    def fresnel_reflection(
        self,
        material2,
        angle_in,
        material1=None,
        forward: bool = True,
        s_polarized: bool = True,
        path: str = "./RefractiveIndices/",
    ):
        """Calculates the fresnel reflection of the pulse at the interface between two materials. The waveform is travelling from material 1 to material 2.
        The first medium (material1) should be non-absorptive. As usual the resulting waveform is stored in the TraceHandler object, replacing the previous one.

        Refractive index files should contain 3 space-separated columns, respectively with headers: wvl n k, where wvl is the wavelength in um, n and k resp. the real and imaginary part of the refractive index.

        Args:
            material2: the filename (without '.txt') of the refractive index data for the material after the interface (e.g. Si Al MgF2); wavelength is in um in the refractive index file
            angle_in: the incidence angle in degrees
            material1: the filename (without '.txt') of the refractive index data for the material before the interface. If None (default), vacuum is assumed
            forward (bool):  if True (default) forward reflection is computed (the result waveform is the reflection of the previously stored waveform.

                if False backward reflection is computed (the previous waveform is the reflection of the result waveform)
            s_polarized (bool): True. Reflection calculation only implemented for s-polarized light
            path (str): path for the refractive index files. Defaults to "./RefractiveIndices/"

        """
        if not s_polarized:
            raise ValueError(
                "in function TraceHandler.fresnel_reflection() p_polarized is not implemented yet\n"
            )

        # read refractive index
        refIndexData = pd.read_table(
            path + material2 + ".txt", sep=" ", keep_default_na=False
        )
        wvl2 = np.array(refIndexData["wvl"]) * 1e3
        n2 = np.array(refIndexData["n"]) + 1j * np.array(refIndexData["k"])
        if material1 is None:
            wvl1 = np.array(wvl2)
            n1 = n2 * 0 + 1
        else:
            refIndexData = pd.read_table(
                path + material1 + ".txt", sep=" ", keep_default_na=False
            )
            wvl1 = np.array(refIndexData["wvl"]) * 1e3
            n1 = np.array(refIndexData["n"]) + 1j * np.array(refIndexData["k"])

        # check that the wvl arrays are monotonically increasing
        if np.any(np.diff(wvl1) < 0):
            if np.all(np.diff(wvl1) < 0):
                wvl1 = wvl1[::-1]
                n1 = n1[::-1]
            else:
                raise ValueError(
                    "in function TraceHandler.fresnel_reflection() wavelength array is not monotonous"
                )
        if np.any(np.diff(wvl2) < 0):
            if np.all(np.diff(wvl2) < 0):
                wvl2 = wvl2[::-1]
                n2 = n2[::-1]
            else:
                raise ValueError(
                    "in function TraceHandler.fresnel_reflection() wavelength array is not monotonous"
                )

        # frequency spectrum (monotonically increasing)
        freq1 = constants.speed_of_light / wvl1[::-1] * 1e-6
        freq2 = constants.speed_of_light / wvl2[::-1] * 1e-6
        n1 = n1[::-1]
        n2 = n2[::-1]

        # fill with ones
        initial_ones_freq = np.linspace(
            freq1[1] - freq1[0],
            freq1[0],
            int(np.ceil(4 * freq1[0] / (freq1[1] - freq1[0]))),
        )
        initial_ones = np.ones(len(initial_ones_freq)) * n1[0]
        final_ones_freq = np.linspace(
            2 * freq1[-1] - freq1[-2],
            4 * freq1[-1],
            int(np.ceil(15 * freq1[-1] / (freq1[-1] - freq1[-2]))),
        )
        final_ones = np.ones(len(final_ones_freq)) * n1[-1]
        freq1 = np.concatenate((initial_ones_freq, freq1, final_ones_freq))
        n1 = np.concatenate((initial_ones, n1, final_ones))
        # same for n2
        initial_ones_freq = np.linspace(
            freq2[1] - freq2[0],
            freq2[0],
            int(np.ceil(4 * freq2[0] / (freq2[1] - freq2[0]))),
        )
        initial_ones = np.ones(len(initial_ones_freq)) * n2[0]
        final_ones_freq = np.linspace(
            2 * freq2[-1] - freq2[-2],
            4 * freq2[-1],
            int(np.ceil(15 * freq2[-1] / (freq2[-1] - freq2[-2]))),
        )
        final_ones = np.ones(len(final_ones_freq)) * n2[-1]
        freq2 = np.concatenate((initial_ones_freq, freq2, final_ones_freq))
        n2 = np.concatenate((initial_ones, n2, final_ones))

        # add negative frequencies
        freq1 = np.concatenate((-freq1[::-1], freq1))
        n1 = np.concatenate((np.conjugate(n1[::-1]), n1))
        freq2 = np.concatenate((-freq2[::-1], freq2))
        n2 = np.concatenate((np.conjugate(n2[::-1]), n2))

        # interpolate the n1 and n2 to the frequency axis of the fft
        n1Interp = np.interp(self.frequencyAxis, freq1, n1)
        n2Interp = np.interp(self.frequencyAxis, freq2, n2)

        # calculate the fresnel reflection coefficients using only the angle of incidence (NOT SURE THIS WORKS WHEN THE FIRST MEDIUM IS LOSSY)
        r = (
            n1Interp * np.cos(angle_in * np.pi / 180)
            - n2Interp
            * np.sqrt(1 - (n1Interp / n2Interp * np.sin(angle_in * np.pi / 180)) ** 2)
        ) / (
            n1Interp * np.cos(angle_in * np.pi / 180)
            + n2Interp
            * np.sqrt(1 - (n1Interp / n2Interp * np.sin(angle_in * np.pi / 180)) ** 2)
        )
        if forward:
            self.fftFieldV = self.fftFieldV * r
        else:
            self.fftFieldV = self.fftFieldV / r

        self.fieldStdevV = None

        self.update_trace_from_fft()
        self.strip_from_trace()
        self.update_fft_spectrum()

    def apply_zero_phase(self):
        """Applies zero-phase to the trace; this allows, for example, to retrieve the fourier limited pulse corresponding to the same FFT spectrum of the loaded trace."""
        tau = (
            self.fsZeroPadding + (np.max(self.fieldTimeV) - np.min(self.fieldTimeV)) / 2
        )
        self.fftFieldV = np.abs(self.fftFieldV) * np.exp(
            1.0j * (2 * np.pi * tau * self.frequencyAxis)
        )
        self.update_trace_from_fft()
        self.strip_from_trace()
        self.update_fft_spectrum()

    def time_frequency_analysis(
        self, sigma_time, low_lim=None, up_lim=None, low_lim_freq=None, up_lim_freq=None
    ):
        """Performs time-frequency analysis by using scipy's short time fourier transform (fourier transform of the trace convoluted by a 'sigma_time' broad gaussian).

        Args:
            sigma_time: sigma of the gaussian window
            low_lim, up_lim (float): xaxis limits for plotting. Default None
            low_lim_freq, up_lim_freq (float): xaxis limits for plotting. Default None

        """
        dt = np.mean(np.diff(self.fieldTimeV))
        w = scipy.signal.windows.gaussian(
            int(sigma_time / dt * 6) + 1, sigma_time / dt, sym=True
        )
        TFA = scipy.signal.ShortTimeFFT(
            w, hop=1, fs=1.0 / dt, mfft=int(sigma_time / dt * 24), scale_to="magnitude"
        )
        TFData = TFA.stft(self.fieldV)

        fig, ax = plt.subplots()
        t_lo, t_hi, f_lo, f_hi = TFA.extent(
            self.fieldV.size
        )  # time and freq range of plot
        if low_lim is None:
            low_lim = t_lo + self.fieldTimeV[0]
        if up_lim is None:
            up_lim = t_hi + self.fieldTimeV[0]
        if low_lim_freq is None:
            low_lim_freq = 0
        if up_lim_freq is None:
            up_lim_freq = 3.0

        ax.set(xlim=(low_lim, up_lim), ylim=(low_lim_freq, up_lim_freq))
        ax.set_xlabel("Time (fs)")
        ax.set_ylabel("Frequency (PHz)")
        im1 = ax.imshow(
            abs(TFData) / np.max(abs(TFData)),
            origin="lower",
            aspect="auto",
            extent=(t_lo + self.fieldTimeV[0], t_hi + self.fieldTimeV[0], f_lo, f_hi),
            cmap="viridis",
        )
        cbar = fig.colorbar(im1)

        cbar.ax.set_ylabel("Magnitude of the field (Arb. unit)")
        fig.tight_layout()
        return TFData

    def plot_trace(self, low_lim=None, up_lim=None, normalize: bool = True):
        """Plots the field trace.

        Args:
            low_lim, up_lim: float = xaxis limits for plotting. Default None
            normalize: bool = if True (default) normalize the peak of the trace to 1

        """
        fig, ax = plt.subplots()
        if normalize:
            norm_plot = self.normalization_trace
        else:
            norm_plot = 1
        main_line = ax.plot(
            self.fieldTimeV, self.fieldV / norm_plot, label="Field trace"
        )
        if self.fieldStdevV is not None:
            ax.fill_between(
                self.fieldTimeV,
                (self.fieldV - self.fieldStdevV) / norm_plot,
                (self.fieldV + self.fieldStdevV) / norm_plot,
                color=main_line[0].get_color(),
                alpha=0.3,
            )
        ax.set_xlabel("Time (fs)")
        ax.set_ylabel("Field (Arb. unit)")

        if low_lim is not None and up_lim is not None:
            ax.set_xlim(low_lim, up_lim)
        return fig

    def plot_spectrum(
        self, low_lim=40, up_lim=1000, no_phase: bool = False, phase_blanking_level=0.05
    ):
        """Plots the trace spectrum and phase together with the spectrometer measurement [if provided].

        Args:
            low_lim, up_lim (float): xaxis limits for plotting. Default: 40, 1000
            no_phase: if True, don't plot the phase. Default: False

        """
        fig, ax = plt.subplots()

        if not no_phase:
            ax2 = ax.twinx()
        lines = []
        min_intensity = phase_blanking_level * np.max(self.fftSpectrum)
        lines += ax.plot(
            self.wvlAxis[(self.wvlAxis > low_lim) & (self.wvlAxis < up_lim)],
            self.fftSpectrum[(self.wvlAxis > low_lim) & (self.wvlAxis < up_lim)],
            label="Fourier transform",
        )
        if not no_phase:
            ax2.plot([], [])
            if self.wvlSpectrometer is not None:
                ax2.plot([], [])
            lines += ax2.plot(
                self.wvlAxis[
                    (self.wvlAxis > low_lim)
                    & (self.wvlAxis < up_lim)
                    & (self.fftSpectrum > min_intensity)
                ],
                self.fftphase[
                    (self.wvlAxis > low_lim)
                    & (self.wvlAxis < up_lim)
                    & (self.fftSpectrum > min_intensity)
                ],
                "--",
                label="Phase",
            )
        if self.wvlSpectrometer is not None:
            lines += ax.plot(
                self.wvlSpectrometer[
                    (self.wvlSpectrometer > low_lim) & (self.wvlSpectrometer < up_lim)
                ],
                self.ISpectrometer[
                    (self.wvlSpectrometer > low_lim) & (self.wvlSpectrometer < up_lim)
                ],
                label="Spectrometer",
            )
        ax.set_xlabel("Wavelength (nm)")
        ax.set_ylabel("Intensity (Arb. unit)")
        ax.tick_params(axis="both")
        if not no_phase:
            ax2.set_ylabel("Phase (rad)")
            ax2.tick_params(axis="y")
        ax.legend(lines, [line.get_label() for line in lines])
        return fig


class MultiTraceHandler:
    """Initializes and stores multiple TraceHandler objects. It is used to plot multiple traces in a single graph, compare them, and perform operations between them.

    ALL TIMES IN FS, ALL WAVELENGTHS IN NM, ALL FREQUENCIES IN PHz

    All stored traces are accessible as TraceHandler object in the corresponding list.

    Attributes:
        traceHandlers (list): list of traceHandler objects
        fsZeroPadding (float): zero padding in fs for the traces (default 150 fs)

    """

    def __init__(
        self,
        filenameList=None,
        filenameSpectrumList=None,
        timeList=None,
        fieldList=None,
        stdevList=None,
        wvlList=None,
        spectrumList=None,
        traceHandlerList=None,
    ):
        """Constructor of MultiTraceHandler.

        One of the following arguments must be provided (all arguments default to None:
        * filenameList: list of filenames containing the field traces (see TraceHandler constructor)
        * timeList AND fieldList: list of time arrays and field arrays (see TraceHandler constructor)
        * traceHandlerList: list of TraceHandler objects

        Args:
            filenameList : list of filenames containing the field traces (see TraceHandler constructor)
            filenameSpectrumList : list of filenames containing the spectrum traces (spectrometer data; see TraceHandler constructor)
            timeList : list of time arrays (see TraceHandler constructor)
            fieldList : list of field arrays (see TraceHandler constructor)
            stdevList : list of standard deviation arrays (see TraceHandler constructor)
            wvlList : list of wavelength arrays (spectrometer data; see TraceHandler constructor)
            spectrumList : list of spectrum arrays (spectrometer data; see TraceHandler constructor)
            traceHandlerList : list of TraceHandler objects

        """
        self.fsZeroPadding = 150
        self.traceHandlers = []
        if filenameList is not None:
            for i in range(len(filenameList)):
                if filenameSpectrumList is not None:
                    self.traceHandlers.append(
                        TraceHandler(
                            filename=filenameList[i],
                            filename_spectrum=filenameSpectrumList[i],
                        )
                    )
                else:
                    self.traceHandlers.append(TraceHandler(filename=filenameList[i]))
        elif timeList is not None and fieldList is not None:
            if stdevList is None:
                stdevList = [None] * len(timeList)
            if wvlList is None:
                wvlList = [None] * len(timeList)
            if spectrumList is None:
                spectrumList = [None] * len(timeList)
            check_equal_length(timeList, fieldList, stdevList, wvlList, spectrumList)
            for i in range(len(timeList)):
                self.traceHandlers.append(
                    TraceHandler(
                        time=timeList[i],
                        field=fieldList[i],
                        stdev=stdevList[i],
                        wvl=wvlList[i],
                        spectrum=spectrumList[i],
                    )
                )
        elif traceHandlerList is not None:
            for i in range(len(traceHandlerList)):
                if isinstance(traceHandlerList[i], TraceHandler):
                    self.traceHandlers.append(traceHandlerList[i])
                else:
                    raise ValueError(
                        "in function MultiTraceHandler.__init__() traceHandlerList is not a list of TraceHandler objects\n"
                    )
        else:
            raise ValueError(
                "in function MultiTraceHandler.__init__() too many arguments were None\n"
                "probably the constructor was erroneously used with no or too few arguments\n"
                "use either filenameList and filenameSpectrumList or timeList, fieldList or traceHandlerList\n"
            )

    def append_trace(
        self,
        filename=None,
        filename_spectrum=None,
        timeV=None,
        fieldV=None,
        stdevV=None,
        wvl=None,
        spectrum=None,
        traceHandler=None,
    ):
        """Append a new trace to the list. Usual rules apply.

        Args:
            filename: file path
            filename_spectrum: spectrum path
            timeV: time vector
            fieldV: field vector
            stdevV: standard deviation
            wvl: wavelength
            spectrum: spectrum
            traceHandler: TraceHandler instance

        """
        if filename_spectrum is not None and filename is not None:
            self.traceHandlers.append(TraceHandler(filename, filename_spectrum))
        elif filename is not None:
            self.traceHandlers.append(TraceHandler(filename))
        elif timeV is not None and fieldV is not None:
            self.traceHandlers.append(
                TraceHandler(
                    time=timeV, field=fieldV, stdev=stdevV, wvl=wvl, spectrum=spectrum
                )
            )
        elif traceHandler is not None:
            if isinstance(traceHandler, TraceHandler):
                self.traceHandlers.append(traceHandler)
            else:
                raise ValueError(
                    "in function MultiTraceHandler.append_trace() traceHandler is not a TraceHandler object\n"
                )
        else:
            raise ValueError(
                "in function MultiTraceHandler.append_trace() filename and filename_spectrum cannot be None\n"
                "it might be that the function was erroneously used with no arguments\n"
            )

    def flip_trace(self, index: int):
        """Flips the trace number 'index'.

        Args:
            index: int

        """
        if index >= len(self.traceHandlers):
            raise ValueError(
                "in function MultiTraceHandler.flip() index is out of range\n"
            )
        self.traceHandlers[index].fieldV = -self.traceHandlers[index].fieldV
        self.traceHandlers[index].fftFieldV = -self.traceHandlers[index].fftFieldV
        self.traceHandlers[index].complexFieldV = -self.traceHandlers[
            index
        ].complexFieldV

    def tukey_bandpass(
        self, lowWavelengthEdge, upWavelengthEdge, lowEdgeWidth, highEdgeWidth
    ):
        """Applies a bandpass filter the traces in the frequency domain using a tukey window. See traceHandler's docs."""
        for i in range(len(self.traceHandlers)):
            self.traceHandlers[i].fft_tukey_bandpass(
                lowWavelengthEdge, upWavelengthEdge, lowEdgeWidth, highEdgeWidth
            )

    def apply_spectrum(
        self, wvl=None, spectrum=None, CEP_shift=0.0, stripZeroPadding=True
    ):
        """Applies spectrum to the phase of the pulse. See traceHandler's docs."""
        for i in range(len(self.traceHandlers)):
            self.traceHandlers[i].apply_spectrum(
                wvl, spectrum, CEP_shift, stripZeroPadding
            )

    def plot_traces(
        self,
        low_lim=None,
        up_lim=None,
        labels=None,
        delay_shift=None,
        offset: float = 2.0,
        errorbar: bool = False,
        normalize: bool = True,
    ):
        """Plots all traces.

        Args:
            low_lim (float): for xaxis
            up_lim (float): for xaxis
            labels (list): list of labels for the legend
            delay_shift (list): list of delay offsets (one per trace)
            offset (float): y-axis offset between traces
            errorbar (bool): whether to plot errors. Default False
            normalize (bool): Default True

        """
        fig, ax = plt.subplots()
        if delay_shift is None:
            delay_shift = [0.0] * len(self.traceHandlers)
        check_equal_length(delay_shift, self.traceHandlers)

        for i in range(len(self.traceHandlers)):
            if normalize:
                norm_plot = self.traceHandlers[i].normalization_trace
            else:
                norm_plot = 1
            t, field = self.traceHandlers[i].get_trace()
            stdev_field = self.traceHandlers[i].get_stdev()
            if errorbar and stdev_field is not None:
                last_fill = ax.fill_between(
                    t - delay_shift[i],
                    offset * i + (field - stdev_field) / norm_plot,
                    offset * i + (field + stdev_field) / norm_plot,
                    label="_nolegend_",
                    alpha=0.3,
                )
                ax.plot(
                    t - delay_shift[i],
                    offset * i + field / norm_plot,
                    label="Trace " + str(i),
                    color=last_fill.get_facecolor(),
                    alpha=1.0,
                )
            else:
                ax.plot(
                    t - delay_shift[i],
                    offset * i + field / norm_plot,
                    label="Trace " + str(i),
                )
        ax.set_xlabel("Time (fs)")
        ax.set_ylabel("Field (Arb. unit)")

        if labels is not None:
            handles, labels_dump = ax.get_legend_handles_labels()
            plt.legend(handles[::-1], labels[::-1], loc="upper left")
        else:
            pass
        if low_lim is not None and up_lim is not None:
            ax.set_xlim(low_lim, up_lim)
        return fig

    def plot_spectra(
        self, low_lim=50, up_lim=1000, labels=None, offset=0.015, logscale: bool = False
    ):
        """Plot all spectra.
        low_lim: lower limit wavelength (nm). Default: 50 nm
        up_lim: upper limit wavelength (nm). Default: 1000 nm
        labels: label list for the plot legend. Labels should be in the same order as the stored traceHandler objects
        offset: artificial offset between two spectra for display purposes. Default: 0.015
        logscale: (bool); whether to plot in a logscale.

        """
        fig, ax = plt.subplots()
        for i in range(len(self.traceHandlers)):
            self.traceHandlers[i].normalize_fft_spectrum()
            wvl, spctr = self.traceHandlers[i].get_spectrum_trace()
            if logscale:
                ax.plot(wvl, (offset**i) * spctr)
                ylow = np.mean(spctr[(wvl < up_lim) & (wvl > low_lim)]) / (offset**3)
                yup = np.mean(spctr[(wvl < up_lim) & (wvl > low_lim)]) * (
                    offset ** (len(self.traceHandlers))
                )
                ax.set_ylim(ylow, yup)
            else:
                ax.plot(wvl, i * offset + spctr)
        ax.set_xlabel("Wavelength (nm)")
        ax.set_ylabel("Intensity (Arb. unit)")
        if labels is not None:
            handles, labels_dump = ax.get_legend_handles_labels()
            plt.legend(handles[::-1], labels[::-1], loc="upper right")
        else:
            pass
        if low_lim is not None and up_lim is not None:
            ax.set_xlim(low_lim, up_lim)
        return fig



"""A few small functions that I might take apart and put into the main module."""

from cycler import cycler
from matplotlib import rcParams


def dark_plot():
    """Use a dark style for matplotlib plots."""
    rcParams["font.family"] = "sans-serif"
    rcParams["font.sans-serif"] = [
        "Helvetica",
        "Arial",
        "Verdana",
        "DejaVu Sans",
        "Liberation Sans",
        "Bitstream Vera Sans",
        "sans-serif",
    ]
    rcParams["axes.prop_cycle"] = cycler(
        color=["cyan", "magenta", "orange", "blueviolet", "lime"]
    )
    rcParams["figure.facecolor"] = "#171717"
    rcParams["figure.edgecolor"] = rcParams["figure.facecolor"]
    rcParams["savefig.facecolor"] = rcParams["figure.facecolor"]
    rcParams["savefig.edgecolor"] = rcParams["figure.facecolor"]
    rcParams["axes.facecolor"] = "black"
    rcParams["text.color"] = "white"
    rcParams["axes.edgecolor"] = "white"
    rcParams["axes.labelcolor"] = "white"
    rcParams["xtick.color"] = "white"
    rcParams["ytick.color"] = "white"
    rcParams["grid.color"] = "white"
    rcParams["lines.color"] = "white"
